#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3300 - Module for testing Simple Oracle Document Access (SODA) Database
"""

import json
import unittest

import oracledb
import test_env

@unittest.skipIf(test_env.skip_soda_tests(),
                 "unsupported client/server combination")
class TestCase(test_env.BaseTestCase):

    def __drop_existing_collections(self, soda_db):
        for name in soda_db.getCollectionNames():
            soda_db.openCollection(name).drop()

    def __verify_doc(self, doc, raw_content, str_content=None, content=None,
                     key=None, media_type='application/json'):
        self.assertEqual(doc.getContentAsBytes(), raw_content)
        if str_content is not None:
            self.assertEqual(doc.getContentAsString(), str_content)
        if content is not None:
            self.assertEqual(doc.getContent(), content)
        self.assertEqual(doc.key, key)
        self.assertEqual(doc.mediaType, media_type)

    def test_3300_create_document_with_json(self):
        "3300 - test creating documents with JSON data"
        soda_db = self.connection.getSodaDatabase()
        val = {"testKey1": "testValue1", "testKey2": "testValue2"}
        str_val = json.dumps(val)
        bytes_val = str_val.encode()
        key = "MyKey"
        media_type = "text/plain"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, bytes_val, str_val, val)
        doc = soda_db.createDocument(str_val, key)
        self.__verify_doc(doc, bytes_val, str_val, val, key)
        doc = soda_db.createDocument(bytes_val, key, media_type)
        self.__verify_doc(doc, bytes_val, str_val, val, key, media_type)

    def test_3301_create_document_with_raw(self):
        "3301 - test creating documents with raw data"
        soda_db = self.connection.getSodaDatabase()
        val = b"<html/>"
        key = "MyRawKey"
        media_type = "text/html"
        doc = soda_db.createDocument(val)
        self.__verify_doc(doc, val)
        doc = soda_db.createDocument(val, key)
        self.__verify_doc(doc, val, key=key)
        doc = soda_db.createDocument(val, key, media_type)
        self.__verify_doc(doc, val, key=key, media_type=media_type)

    def test_3302_get_collection_names(self):
        "3302 - test getting collection names from the database"
        soda_db = self.connection.getSodaDatabase()
        self.__drop_existing_collections(soda_db)
        self.assertEqual(soda_db.getCollectionNames(), [])
        names = ["zCol", "dCol", "sCol", "aCol", "gCol"]
        sorted_names = list(sorted(names))
        for name in names:
            soda_db.createCollection(name)
        self.assertEqual(soda_db.getCollectionNames(), sorted_names)
        self.assertEqual(soda_db.getCollectionNames(limit=2), sorted_names[:2])
        self.assertEqual(soda_db.getCollectionNames("a"), sorted_names)
        self.assertEqual(soda_db.getCollectionNames("C"), sorted_names)
        self.assertEqual(soda_db.getCollectionNames("b", limit=3),
                                                    sorted_names[1:4])
        self.assertEqual(soda_db.getCollectionNames("z"), sorted_names[-1:])

    def test_3303_open_collection(self):
        "3303 - test opening a collection"
        soda_db = self.connection.getSodaDatabase()
        self.__drop_existing_collections(soda_db)
        coll = soda_db.openCollection("CollectionThatDoesNotExist")
        self.assertEqual(coll, None)
        created_coll = soda_db.createCollection("TestOpenCollection")
        coll = soda_db.openCollection(created_coll.name)
        self.assertEqual(coll.name, created_coll.name)
        coll.drop()

    def test_3304_repr(self):
        "3304 - test SodaDatabase representation"
        con1 = self.connection
        con2 = test_env.get_connection()
        soda_db1 = self.connection.getSodaDatabase()
        soda_db2 = con1.getSodaDatabase()
        soda_db3 = con2.getSodaDatabase()
        self.assertEqual(str(soda_db1), str(soda_db2))
        self.assertEqual(str(soda_db2), str(soda_db3))

    def test_3305_negative(self):
        "3305 - test negative cases for SODA database methods"
        soda_db = self.connection.getSodaDatabase()
        self.assertRaises(TypeError, soda_db.createCollection)
        self.assertRaises(TypeError, soda_db.createCollection, 1)
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-40658:",
                               soda_db.createCollection, None)
        self.assertRaises(TypeError, soda_db.getCollectionNames, 1)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# buffer.pyx
#
# Cython file defining the StringBuffer class, used for transforming Python
# "str" and "bytes" objects into the C buffer representation required by ODPI-C
# (embedded in thick_impl.pyx).
#------------------------------------------------------------------------------

@cython.freelist(20)
cdef class StringBuffer:
    cdef:
        bytes obj
        char *ptr
        uint32_t length
        uint32_t size_in_chars

    cdef int set_value(self, value) except -1:
        if value is None:
            self.obj = None
            self.ptr = NULL
            self.length = self.size_in_chars = 0
            return 0
        elif isinstance(value, str):
            self.obj = (<str> value).encode()
            self.size_in_chars = <uint32_t> len(<str> value)
        elif isinstance(value, bytes):
            self.obj = value
            self.size_in_chars = <uint32_t> len(<bytes> value)
        else:
            raise TypeError("expecting string or bytes")
        self.ptr = self.obj
        self.length = <uint32_t> len(self.obj)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# odpi.pxd
#
# Cython definition file for the constants, functions and structures found in
# ODPI-C used by the thick implementation classes (embedded in thick_impl.pyx).
#------------------------------------------------------------------------------

from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t

cdef enum:
    PYO_OCI_ATTR_TYPE_STRING = 1
    PYO_OCI_ATTR_TYPE_BOOLEAN = 2
    PYO_OCI_ATTR_TYPE_UINT8 = 8
    PYO_OCI_ATTR_TYPE_UINT16 = 16
    PYO_OCI_ATTR_TYPE_UINT32 = 32
    PYO_OCI_ATTR_TYPE_UINT64 = 64

cdef extern from "impl/thick/odpi/embed/dpi.c":

    # version constants
    unsigned int DPI_MAJOR_VERSION
    unsigned int DPI_MINOR_VERSION

    # success/failure constants
    int DPI_SUCCESS
    int DPI_FAILURE

    # defaults
    enum:
        DPI_DEFAULT_PREFETCH_ROWS

    # execution modes
    enum:
        DPI_MODE_EXEC_ARRAY_DML_ROWCOUNTS
        DPI_MODE_EXEC_BATCH_ERRORS
        DPI_MODE_EXEC_COMMIT_ON_SUCCESS
        DPI_MODE_EXEC_DEFAULT
        DPI_MODE_EXEC_DESCRIBE_ONLY
        DPI_MODE_EXEC_PARSE_ONLY

    # connection/pool creation modes
    enum:
        DPI_MODE_CREATE_DEFAULT
        DPI_MODE_CREATE_THREADED
        DPI_MODE_CREATE_EVENTS

    # connection close modes
    enum:
        DPI_MODE_CONN_CLOSE_DEFAULT
        DPI_MODE_CONN_CLOSE_DROP
        DPI_MODE_CONN_CLOSE_RETAG

    # pool close modes
    enum:
        DPI_MODE_POOL_CLOSE_DEFAULT
        DPI_MODE_POOL_CLOSE_FORCE

    # native types
    enum:
        DPI_NATIVE_TYPE_BOOLEAN
        DPI_NATIVE_TYPE_BYTES
        DPI_NATIVE_TYPE_DOUBLE
        DPI_NATIVE_TYPE_FLOAT
        DPI_NATIVE_TYPE_INT64
        DPI_NATIVE_TYPE_INTERVAL_DS
        DPI_NATIVE_TYPE_JSON
        DPI_NATIVE_TYPE_JSON_ARRAY
        DPI_NATIVE_TYPE_JSON_OBJECT
        DPI_NATIVE_TYPE_LOB
        DPI_NATIVE_TYPE_NULL
        DPI_NATIVE_TYPE_OBJECT
        DPI_NATIVE_TYPE_ROWID
        DPI_NATIVE_TYPE_STMT
        DPI_NATIVE_TYPE_TIMESTAMP

    # Oracle types
    enum:
        DPI_ORACLE_TYPE_BFILE
        DPI_ORACLE_TYPE_BLOB
        DPI_ORACLE_TYPE_BOOLEAN
        DPI_ORACLE_TYPE_CHAR
        DPI_ORACLE_TYPE_CLOB
        DPI_ORACLE_TYPE_DATE
        DPI_ORACLE_TYPE_INTERVAL_DS
        DPI_ORACLE_TYPE_INTERVAL_YM
        DPI_ORACLE_TYPE_JSON
        DPI_ORACLE_TYPE_JSON_ARRAY
        DPI_ORACLE_TYPE_JSON_OBJECT
        DPI_ORACLE_TYPE_LONG_NVARCHAR
        DPI_ORACLE_TYPE_LONG_RAW
        DPI_ORACLE_TYPE_LONG_VARCHAR
        DPI_ORACLE_TYPE_NATIVE_DOUBLE
        DPI_ORACLE_TYPE_NATIVE_FLOAT
        DPI_ORACLE_TYPE_NATIVE_INT
        DPI_ORACLE_TYPE_NCHAR
        DPI_ORACLE_TYPE_NCLOB
        DPI_ORACLE_TYPE_NONE
        DPI_ORACLE_TYPE_NUMBER
        DPI_ORACLE_TYPE_NVARCHAR
        DPI_ORACLE_TYPE_OBJECT
        DPI_ORACLE_TYPE_RAW
        DPI_ORACLE_TYPE_ROWID
        DPI_ORACLE_TYPE_STMT
        DPI_ORACLE_TYPE_TIMESTAMP
        DPI_ORACLE_TYPE_TIMESTAMP_LTZ
        DPI_ORACLE_TYPE_TIMESTAMP_TZ
        DPI_ORACLE_TYPE_VARCHAR

    # SODA flags
    enum:
        DPI_SODA_FLAGS_DEFAULT
        DPI_SODA_FLAGS_ATOMIC_COMMIT
        DPI_SODA_FLAGS_CREATE_COLL_MAP
        DPI_SODA_FLAGS_INDEX_DROP_FORCE

    # database startup modes
    enum:
        DPI_MODE_STARTUP_DEFAULT
        DPI_MODE_STARTUP_FORCE
        DPI_MODE_STARTUP_RESTRICT

    # fetch modes (for scrolling cursors)
    enum:
        DPI_MODE_FETCH_FIRST
        DPI_MODE_FETCH_LAST
        DPI_MODE_FETCH_ABSOLUTE
        DPI_MODE_FETCH_RELATIVE

    # subscription constants
    uint32_t DPI_SUBSCR_QOS_QUERY
    uint32_t DPI_EVENT_OBJCHANGE
    uint32_t DPI_EVENT_QUERYCHANGE

    # JSON flags
    uint32_t DPI_JSON_OPT_NUMBER_AS_STRING

    # opaque handles
    ctypedef struct dpiConn:
        pass

    ctypedef struct dpiContext:
        pass

    ctypedef struct dpiDeqOptions:
        pass

    ctypedef struct dpiEnqOptions:
        pass

    ctypedef struct dpiJson:
        pass

    ctypedef struct dpiLob:
        pass

    ctypedef struct dpiMsgProps:
        pass

    ctypedef struct dpiObject:
        pass

    ctypedef struct dpiObjectAttr:
        pass

    ctypedef struct dpiObjectType:
        pass

    ctypedef struct dpiPool:
        pass

    ctypedef struct dpiQueue:
        pass

    ctypedef struct dpiRowid:
        pass

    ctypedef struct dpiSodaColl:
        pass

    ctypedef struct dpiSodaDb:
        pass

    ctypedef struct dpiSodaDoc:
        pass

    ctypedef struct dpiSodaDocCursor:
        pass

    ctypedef struct dpiStmt:
        pass

    ctypedef struct dpiSubscr:
        pass

    ctypedef struct dpiVar:
        pass

    # function pointer types
    ctypedef void (*dpiSubscrCallback)(void* context,
            dpiSubscrMessage *message)

    ctypedef int (*dpiAccessTokenCallback)(void *context,
            dpiAccessToken *accessToken)

    # complex native data types
    ctypedef struct dpiBytes:
        char *ptr
        uint32_t length
        const char *encoding

    ctypedef struct dpiIntervalDS:
        int32_t days
        int32_t hours
        int32_t minutes
        int32_t seconds
        int32_t fseconds

    ctypedef struct dpiIntervalYM:
        int32_t years
        int32_t months

    ctypedef struct dpiJsonNode:
        uint32_t oracleTypeNum
        uint32_t nativeTypeNum
        dpiDataBuffer *value

    ctypedef struct dpiJsonArray:
        uint32_t numElements
        dpiJsonNode *elements
        dpiDataBuffer *elementValues

    ctypedef struct dpiJsonObject:
        uint32_t numFields
        char **fieldNames
        uint32_t *fieldNameLengths
        dpiJsonNode *fields
        dpiDataBuffer *fieldValues

    ctypedef struct dpiTimestamp:
        int16_t year
        uint8_t month
        uint8_t day
        uint8_t hour
        uint8_t minute
        uint8_t second
        uint32_t fsecond
        int8_t tzHourOffset
        int8_t tzMinuteOffset

    # public structures
    ctypedef struct dpiAppContext:
        const char *namespaceName
        uint32_t namespaceNameLength
        const char *name
        uint32_t nameLength
        const char *value
        uint32_t valueLength

    ctypedef struct dpiCommonCreateParams:
        uint32_t createMode
        const char *encoding
        const char *nencoding
        const char *edition
        uint32_t editionLength
        const char *driverName
        uint32_t driverNameLength
        bint sodaMetadataCache
        uint32_t stmtCacheSize
        dpiAccessToken *accessToken

    ctypedef struct dpiConnCreateParams:
        uint32_t authMode
        const char *connectionClass
        uint32_t connectionClassLength
        uint32_t purity
        const char *newPassword
        uint32_t newPasswordLength
        dpiAppContext *appContext
        uint32_t numAppContext
        bint externalAuth
        void *externalHandle
        dpiPool *pool
        const char *tag
        uint32_t tagLength
        bint matchAnyTag
        const char *outTag
        uint32_t outTagLength
        bint outTagFound
        dpiShardingKeyColumn *shardingKeyColumns
        uint8_t numShardingKeyColumns
        dpiShardingKeyColumn *superShardingKeyColumns
        uint8_t numSuperShardingKeyColumns
        bint outNewSession

    ctypedef struct dpiContextCreateParams:
        const char *defaultDriverName
        const char *defaultEncoding
        const char *loadErrorUrl
        const char *oracleClientLibDir
        const char *oracleClientConfigDir

    ctypedef union dpiDataBuffer:
        bint asBoolean
        uint8_t asUint8
        uint16_t asUint16
        uint32_t asUint32
        int64_t asInt64
        uint64_t asUint64
        float asFloat
        double asDouble
        char *asString
        void *asRaw
        dpiBytes asBytes
        dpiTimestamp asTimestamp
        dpiIntervalDS asIntervalDS
        dpiIntervalYM asIntervalYM
        dpiJson *asJson
        dpiJsonObject asJsonObject
        dpiJsonArray asJsonArray
        dpiLob *asLOB
        dpiObject *asObject
        dpiStmt *asStmt
        dpiRowid *asRowid

    ctypedef struct dpiData:
        bint isNull
        dpiDataBuffer value

    ctypedef struct dpiDataTypeInfo:
        uint32_t oracleTypeNum
        uint32_t defaultNativeTypeNum
        uint16_t ociTypeCode
        uint32_t dbSizeInBytes
        uint32_t clientSizeInBytes
        uint32_t sizeInChars
        int16_t precision
        int8_t scale
        uint8_t fsPrecision
        dpiObjectType *objectType

    ctypedef struct dpiAccessToken:
        const char *token
        uint32_t tokenLength
        const char *privateKey
        uint32_t privateKeyLength

    ctypedef struct dpiErrorInfo:
        int32_t code
        uint16_t offset16
        const char *message
        uint32_t messageLength
        const char *encoding
        const char *fnName
        const char *action
        const char *sqlState
        bint isRecoverable
        bint isWarning
        uint32_t offset

    ctypedef struct dpiObjectAttrInfo:
        const char *name
        uint32_t nameLength
        dpiDataTypeInfo typeInfo

    ctypedef struct dpiObjectTypeInfo:
        const char *schema
        uint32_t schemaLength
        const char *name
        uint32_t nameLength
        bint isCollection
        dpiDataTypeInfo elementTypeInfo
        uint16_t numAttributes
        const char *packageName
        uint32_t packageNameLength

    ctypedef struct dpiPoolCreateParams:
        uint32_t minSessions
        uint32_t maxSessions
        uint32_t sessionIncrement
        int pingInterval
        int pingTimeout
        bint homogeneous
        bint externalAuth
        uint32_t getMode
        const char *outPoolName
        uint32_t outPoolNameLength
        uint32_t timeout
        uint32_t waitTimeout
        uint32_t maxLifetimeSession
        const char *plsqlFixupCallback
        uint32_t plsqlFixupCallbackLength
        uint32_t maxSessionsPerShard
        dpiAccessTokenCallback accessTokenCallback
        void *accessTokenCallbackContext

    ctypedef struct dpiQueryInfo:
        const char *name
        uint32_t nameLength
        dpiDataTypeInfo typeInfo
        bint nullOk

    ctypedef struct dpiShardingKeyColumn:
        uint32_t oracleTypeNum
        uint32_t nativeTypeNum
        dpiDataBuffer value

    ctypedef struct dpiSodaCollNames:
        uint32_t numNames
        const char **names
        uint32_t *nameLengths

    ctypedef struct dpiSodaOperOptions:
        uint32_t numKeys
        const char **keys
        uint32_t *keyLengths
        const char *key
        uint32_t keyLength
        const char *version
        uint32_t versionLength
        const char *filter
        uint32_t filterLength
        const char *hint;
        uint32_t hintLength;
        uint32_t skip
        uint32_t limit
        uint32_t fetchArraySize

    ctypedef struct dpiStmtInfo:
        bint isQuery
        bint isPLSQL
        bint isDDL
        bint isDML
        uint16_t statementType
        bint isReturning

    ctypedef struct dpiSubscrCreateParams:
        uint32_t subscrNamespace
        uint32_t protocol
        uint32_t qos
        uint32_t operations
        uint32_t portNumber
        uint32_t timeout
        const char *name
        uint32_t nameLength
        dpiSubscrCallback callback
        void *callbackContext
        const char *recipientName
        uint32_t recipientNameLength
        const char *ipAddress
        uint32_t ipAddressLength
        uint8_t groupingClass
        uint32_t groupingValue
        uint8_t groupingType
        uint64_t outRegId
        bint clientInitiated

    ctypedef struct dpiSubscrMessage:
        uint32_t eventType
        const char *dbName
        uint32_t dbNameLength
        dpiSubscrMessageTable *tables
        uint32_t numTables
        dpiSubscrMessageQuery *queries
        uint32_t numQueries
        dpiErrorInfo *errorInfo
        const void *txId
        uint32_t txIdLength
        bint registered
        const char *queueName
        uint32_t queueNameLength
        const char *consumerName
        uint32_t consumerNameLength
        const void *aqMsgId
        uint32_t aqMsgIdLength

    ctypedef struct dpiSubscrMessageQuery:
        uint64_t id
        uint32_t operation
        dpiSubscrMessageTable *tables
        uint32_t numTables

    ctypedef struct dpiSubscrMessageRow:
        uint32_t operation
        const char *rowid
        uint32_t rowidLength

    ctypedef struct dpiSubscrMessageTable:
        uint32_t operation
        const char *name
        uint32_t nameLength
        dpiSubscrMessageRow *rows
        uint32_t numRows

    ctypedef struct dpiVersionInfo:
        int versionNum
        int releaseNum
        int updateNum
        int portReleaseNum
        int portUpdateNum
        uint32_t fullVersionNum

    ctypedef struct dpiXid:
        long formatId
        const char *globalTransactionId
        uint32_t globalTransactionIdLength
        const char *branchQualifier
        uint32_t branchQualifierLength

    ctypedef struct dpiMsgRecipient:
        const char *name
        uint32_t nameLength

    # functions
    int dpiConn_breakExecution(dpiConn *conn) nogil

    int dpiConn_changePassword(dpiConn *conn, const char *userName,
            uint32_t userNameLength, const char *oldPassword,
            uint32_t oldPasswordLength, const char *newPassword,
            uint32_t newPasswordLength) nogil

    int dpiConn_close(dpiConn *conn, uint32_t mode, const char *tag,
            uint32_t tagLength) nogil

    int dpiConn_commit(dpiConn *conn) nogil

    int dpiConn_create(const dpiContext *context, const char *userName,
            uint32_t userNameLength, const char *password,
            uint32_t passwordLength, const char *connectString,
            uint32_t connectStringLength,
            const dpiCommonCreateParams *commonParams,
            dpiConnCreateParams *createParams, dpiConn **conn) nogil

    int dpiConn_getCurrentSchema(dpiConn *conn, const char **value,
            uint32_t *valueLength) nogil

    int dpiConn_getEdition(dpiConn *conn, const char **value,
            uint32_t *valueLength) nogil

    int dpiConn_getExternalName(dpiConn *conn, const char **value,
            uint32_t *valueLength) nogil

    int dpiConn_getHandle(dpiConn *conn, void **handle) nogil

    int dpiConn_getInternalName(dpiConn *conn, const char **value,
            uint32_t *valueLength) nogil

    int dpiConn_getIsHealthy(dpiConn *conn, bint *isHealthy) nogil

    int dpiConn_getLTXID(dpiConn *conn, const char **value,
            uint32_t *valueLength) nogil

    int dpiConn_getOciAttr(dpiConn *conn, uint32_t handleType,
            uint32_t attribute, dpiDataBuffer *value,
            uint32_t *valueLength) nogil

    int dpiConn_getObjectType(dpiConn *conn, const char *name,
            uint32_t nameLength, dpiObjectType **objType) nogil

    int dpiConn_getServerVersion(dpiConn *conn, const char **releaseString,
            uint32_t *releaseStringLength, dpiVersionInfo *versionInfo) nogil

    int dpiConn_getSodaDb(dpiConn *conn, dpiSodaDb **db) nogil

    int dpiConn_getStmtCacheSize(dpiConn *conn, uint32_t *cacheSize) nogil

    int dpiConn_getCallTimeout(dpiConn *conn, uint32_t *value) nogil

    int dpiConn_newMsgProps(dpiConn *conn, dpiMsgProps **props) nogil

    int dpiConn_newQueue(dpiConn *conn, const char *name,
            uint32_t nameLength, dpiObjectType *payloadType,
            dpiQueue **queue) nogil

    int dpiConn_newJson(dpiConn *conn, dpiJson **json) nogil

    int dpiConn_newJsonQueue(dpiConn *conn, const char *name,
            uint32_t nameLength, dpiQueue **queue) nogil

    int dpiConn_newTempLob(dpiConn *conn, uint32_t lobType, dpiLob **lob) nogil

    int dpiConn_newVar(dpiConn *conn, uint32_t oracleTypeNum,
            uint32_t nativeTypeNum, uint32_t maxArraySize, uint32_t size,
            int sizeIsBytes, bint isArray, dpiObjectType *objType, dpiVar **var,
            dpiData **data) nogil

    int dpiConn_ping(dpiConn *conn) nogil

    int dpiConn_prepareStmt(dpiConn *conn, bint scrollable, const char *sql,
            uint32_t sqlLength, const char *tag, uint32_t tagLength,
            dpiStmt **stmt) nogil

    int dpiConn_release(dpiConn *conn) nogil

    int dpiConn_rollback(dpiConn *conn) nogil

    int dpiConn_setAction(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setCallTimeout(dpiConn *conn, uint32_t value) nogil

    int dpiConn_setClientIdentifier(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setClientInfo(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setCurrentSchema(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setDbOp(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setEcontextId(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setExternalName(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setInternalName(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setModule(dpiConn *conn, const char *value,
            uint32_t valueLength) nogil

    int dpiConn_setOciAttr(dpiConn *conn, uint32_t handleType,
            uint32_t attribute, void *value, uint32_t valueLength) nogil

    int dpiConn_setStmtCacheSize(dpiConn *conn, uint32_t cacheSize) nogil

    int dpiConn_shutdownDatabase(dpiConn *conn, uint32_t mode) nogil

    int dpiConn_startupDatabaseWithPfile(dpiConn *conn, const char *pfile,
            uint32_t pfileLength, uint32_t mode) nogil

    int dpiConn_subscribe(dpiConn *conn, dpiSubscrCreateParams *params,
            dpiSubscr **subscr) nogil

    int dpiConn_tpcBegin(dpiConn *conn, dpiXid *xid,
            uint32_t transactionTimeout, uint32_t flags) nogil

    int dpiConn_tpcCommit(dpiConn *conn, dpiXid *xid, bint onePhase) nogil

    int dpiConn_tpcEnd(dpiConn *conn, dpiXid *xid, uint32_t flags) nogil

    int dpiConn_tpcForget(dpiConn *conn, dpiXid *xid) nogil

    int dpiConn_tpcPrepare(dpiConn *conn, dpiXid *xid,
            bint *commitNeeded) nogil

    int dpiConn_tpcRollback(dpiConn *conn, dpiXid *xid) nogil

    int dpiConn_unsubscribe(dpiConn *conn, dpiSubscr *subscr) nogil

    int dpiContext_createWithParams(unsigned int majorVersion,
            unsigned int minorVersion, dpiContextCreateParams *params,
            dpiContext **context, dpiErrorInfo *errorInfo) nogil

    int dpiContext_getClientVersion(const dpiContext *context,
            dpiVersionInfo *versionInfo) nogil

    void dpiContext_getError(const dpiContext *context,
            dpiErrorInfo *errorInfo) nogil

    int dpiContext_initCommonCreateParams(const dpiContext *context,
            dpiCommonCreateParams *params) nogil

    int dpiContext_initConnCreateParams(const dpiContext *context,
            dpiConnCreateParams *params) nogil

    int dpiContext_initPoolCreateParams(const dpiContext *context,
            dpiPoolCreateParams *params) nogil

    int dpiContext_initSodaOperOptions(const dpiContext *context,
            dpiSodaOperOptions *options) nogil

    int dpiContext_initSubscrCreateParams(const dpiContext *context,
            dpiSubscrCreateParams *params) nogil

    int dpiDeqOptions_getCondition(dpiDeqOptions *options,
            const char **value, uint32_t *valueLength) nogil

    int dpiDeqOptions_getConsumerName(dpiDeqOptions *options,
            const char **value, uint32_t *valueLength) nogil

    int dpiDeqOptions_getCorrelation(dpiDeqOptions *options,
            const char **value, uint32_t *valueLength) nogil

    int dpiDeqOptions_getMode(dpiDeqOptions *options, uint32_t *value) nogil

    int dpiDeqOptions_getMsgId(dpiDeqOptions *options,
            const char **value, uint32_t *valueLength) nogil

    int dpiDeqOptions_getNavigation(dpiDeqOptions *options,
            uint32_t *value) nogil

    int dpiDeqOptions_getTransformation(dpiDeqOptions *options,
            const char **value, uint32_t *valueLength) nogil

    int dpiDeqOptions_getVisibility(dpiDeqOptions *options,
            uint32_t *value) nogil

    int dpiDeqOptions_getWait(dpiDeqOptions *options, uint32_t *value) nogil

    int dpiDeqOptions_release(dpiDeqOptions *options) nogil

    int dpiDeqOptions_setCondition(dpiDeqOptions *options,
            const char *value, uint32_t valueLength) nogil

    int dpiDeqOptions_setConsumerName(dpiDeqOptions *options,
            const char *value, uint32_t valueLength) nogil

    int dpiDeqOptions_setCorrelation(dpiDeqOptions *options,
            const char *value, uint32_t valueLength) nogil

    int dpiDeqOptions_setDeliveryMode(dpiDeqOptions *options,
            uint16_t value) nogil

    int dpiDeqOptions_setMode(dpiDeqOptions *options, uint32_t value) nogil

    int dpiDeqOptions_setMsgId(dpiDeqOptions *options,
            const char *value, uint32_t valueLength) nogil

    int dpiDeqOptions_setNavigation(dpiDeqOptions *options,
            uint32_t value) nogil

    int dpiDeqOptions_setTransformation(dpiDeqOptions *options,
            const char *value, uint32_t valueLength) nogil

    int dpiDeqOptions_setVisibility(dpiDeqOptions *options,
            uint32_t value) nogil

    int dpiDeqOptions_setWait(dpiDeqOptions *options, uint32_t value) nogil

    int dpiEnqOptions_getTransformation(dpiEnqOptions *options,
            const char **value, uint32_t *valueLength) nogil

    int dpiEnqOptions_getVisibility(dpiEnqOptions *options,
            uint32_t *value) nogil

    int dpiEnqOptions_release(dpiEnqOptions *options) nogil

    int dpiEnqOptions_setDeliveryMode(dpiEnqOptions *options,
            uint16_t value) nogil

    int dpiEnqOptions_setTransformation(dpiEnqOptions *options,
            const char *value, uint32_t valueLength) nogil

    int dpiEnqOptions_setVisibility(dpiEnqOptions *options,
            uint32_t value) nogil

    int dpiJson_getValue(dpiJson *json, uint32_t options,
            dpiJsonNode **topNode) nogil

    int dpiJson_setValue(dpiJson *json, dpiJsonNode *topNode)

    int dpiLob_addRef(dpiLob *lob) nogil

    int dpiLob_closeResource(dpiLob *lob) nogil

    int dpiLob_getBufferSize(dpiLob *lob, uint64_t sizeInChars,
            uint64_t *sizeInBytes) nogil

    int dpiLob_getChunkSize(dpiLob *lob, uint32_t *size) nogil

    int dpiLob_getDirectoryAndFileName(dpiLob *lob,
            const char **directoryAlias, uint32_t *directoryAliasLength,
            const char **fileName, uint32_t *fileNameLength) nogil

    int dpiLob_getFileExists(dpiLob *lob, bint *exists) nogil

    int dpiLob_getIsResourceOpen(dpiLob *lob, bint *isOpen) nogil

    int dpiLob_getSize(dpiLob *lob, uint64_t *size) nogil

    int dpiLob_openResource(dpiLob *lob) nogil

    int dpiLob_readBytes(dpiLob *lob, uint64_t offset, uint64_t amount,
            char *value, uint64_t *valueLength) nogil

    int dpiLob_release(dpiLob *lob) nogil

    int dpiLob_setDirectoryAndFileName(dpiLob *lob,
            const char *directoryAlias, uint32_t directoryAliasLength,
            const char *fileName, uint32_t fileNameLength) nogil

    int dpiLob_setFromBytes(dpiLob *lob, const char *value,
            uint64_t valueLength) nogil

    int dpiLob_trim(dpiLob *lob, uint64_t newSize) nogil

    int dpiLob_writeBytes(dpiLob *lob, uint64_t offset,
            const char *value, uint64_t valueLength) nogil

    int dpiMsgProps_getNumAttempts(dpiMsgProps *props, int32_t *value) nogil

    int dpiMsgProps_getCorrelation(dpiMsgProps *props,
            const char **value, uint32_t *valueLength) nogil

    int dpiMsgProps_getDelay(dpiMsgProps *props, int32_t *value) nogil

    int dpiMsgProps_getDeliveryMode(dpiMsgProps *props, uint16_t *value) nogil

    int dpiMsgProps_getEnqTime(dpiMsgProps *props, dpiTimestamp *value) nogil

    int dpiMsgProps_getExceptionQ(dpiMsgProps *props,
            const char **value, uint32_t *valueLength) nogil

    int dpiMsgProps_getExpiration(dpiMsgProps *props, int32_t *value) nogil

    int dpiMsgProps_getMsgId(dpiMsgProps *props, const char **value,
            uint32_t *valueLength) nogil

    int dpiMsgProps_getOriginalMsgId(dpiMsgProps *props,
            const char **value, uint32_t *valueLength) nogil

    int dpiMsgProps_getPayload(dpiMsgProps *props, dpiObject **obj,
            const char **value, uint32_t *valueLength) nogil

    int dpiMsgProps_getPayloadJson(dpiMsgProps *props,
        dpiJson **json) nogil

    int dpiMsgProps_getPriority(dpiMsgProps *props, int32_t *value) nogil

    int dpiMsgProps_getState(dpiMsgProps *props, uint32_t *value) nogil

    int dpiMsgProps_release(dpiMsgProps *props) nogil

    int dpiMsgProps_setCorrelation(dpiMsgProps *props,
            const char *value, uint32_t valueLength) nogil

    int dpiMsgProps_setDelay(dpiMsgProps *props, int32_t value) nogil

    int dpiMsgProps_setExceptionQ(dpiMsgProps *props, const char *value,
            uint32_t valueLength) nogil

    int dpiMsgProps_setExpiration(dpiMsgProps *props, int32_t value) nogil

    int dpiMsgProps_setOriginalMsgId(dpiMsgProps *props,
        const char *value, uint32_t valueLength) nogil

    int dpiMsgProps_setPayloadBytes(dpiMsgProps *props,
            const char *value, uint32_t valueLength) nogil

    int dpiMsgProps_setPayloadObject(dpiMsgProps *props, dpiObject *obj) nogil

    int dpiMsgProps_setPayloadJson(dpiMsgProps *props, dpiJson *json) nogil

    int dpiMsgProps_setPriority(dpiMsgProps *props, int32_t value) nogil

    int dpiMsgProps_setRecipients(dpiMsgProps *props,
            dpiMsgRecipient *recipients, uint32_t numRecipients) nogil

    int dpiObject_addRef(dpiObject *obj) nogil

    int dpiObject_appendElement(dpiObject *obj, uint32_t nativeTypeNum,
            dpiData *value) nogil

    int dpiObject_copy(dpiObject *obj, dpiObject **copiedObj) nogil

    int dpiObject_deleteElementByIndex(dpiObject *obj, int32_t index) nogil

    int dpiObject_getAttributeValue(dpiObject *obj, dpiObjectAttr *attr,
            uint32_t nativeTypeNum, dpiData *value) nogil

    int dpiObject_getElementExistsByIndex(dpiObject *obj, int32_t index,
            bint *exists) nogil

    int dpiObject_getElementValueByIndex(dpiObject *obj, int32_t index,
            uint32_t nativeTypeNum, dpiData *value) nogil

    int dpiObject_getFirstIndex(dpiObject *obj, int32_t *index,
            bint *exists) nogil

    int dpiObject_getLastIndex(dpiObject *obj, int32_t *index,
            bint *exists) nogil

    int dpiObject_getNextIndex(dpiObject *obj, int32_t index,
            int32_t *nextIndex, bint *exists) nogil

    int dpiObject_getPrevIndex(dpiObject *obj, int32_t index,
            int32_t *prevIndex, bint *exists) nogil

    int dpiObject_getSize(dpiObject *obj, int32_t *size) nogil

    int dpiObject_release(dpiObject *obj) nogil

    int dpiObject_setAttributeValue(dpiObject *obj, dpiObjectAttr *attr,
            uint32_t nativeTypeNum, dpiData *value) nogil

    int dpiObject_setElementValueByIndex(dpiObject *obj, int32_t index,
            uint32_t nativeTypeNum, dpiData *value) nogil

    int dpiObject_trim(dpiObject *obj, uint32_t numToTrim) nogil

    int dpiObjectAttr_getInfo(dpiObjectAttr *attr,
            dpiObjectAttrInfo *info) nogil

    int dpiObjectAttr_release(dpiObjectAttr *attr) nogil

    int dpiObjectType_addRef(dpiObjectType *objType) nogil

    int dpiObjectType_createObject(dpiObjectType *objType,
            dpiObject **obj) nogil

    int dpiObjectType_getAttributes(dpiObjectType *objType,
            uint16_t numAttributes, dpiObjectAttr **attributes) nogil

    int dpiObjectType_getInfo(dpiObjectType *objType,
            dpiObjectTypeInfo *info) nogil

    int dpiObjectType_release(dpiObjectType *objType) nogil

    int dpiPool_close(dpiPool *pool, uint32_t closeMode) nogil

    int dpiPool_create(const dpiContext *context, const char *userName,
            uint32_t userNameLength, const char *password,
            uint32_t passwordLength, const char *connectString,
            uint32_t connectStringLength,
            const dpiCommonCreateParams *commonParams,
            dpiPoolCreateParams *createParams, dpiPool **pool) nogil

    int dpiPool_getBusyCount(dpiPool *pool, uint32_t *value) nogil

    int dpiPool_getGetMode(dpiPool *pool, uint8_t *value) nogil

    int dpiPool_getMaxLifetimeSession(dpiPool *pool, uint32_t *value) nogil

    int dpiPool_getMaxSessionsPerShard(dpiPool *pool, uint32_t *value) nogil

    int dpiPool_getOpenCount(dpiPool *pool, uint32_t *value) nogil

    int dpiPool_getPingInterval(dpiPool *pool, int *value) nogil

    int dpiPool_getSodaMetadataCache(dpiPool *pool, bint *enabled) nogil

    int dpiPool_getStmtCacheSize(dpiPool *pool, uint32_t *cacheSize) nogil

    int dpiPool_getTimeout(dpiPool *pool, uint32_t *value) nogil

    int dpiPool_getWaitTimeout(dpiPool *pool, uint32_t *value) nogil

    int dpiPool_release(dpiPool *pool) nogil

    int dpiPool_reconfigure(dpiPool *pool, uint32_t minSessions,
            uint32_t maxSessions, uint32_t sessionIncrement) nogil

    int dpiPool_setAccessToken(dpiPool *pool, dpiAccessToken *params) nogil

    int dpiPool_setGetMode(dpiPool *pool, uint8_t value) nogil

    int dpiPool_setMaxLifetimeSession(dpiPool *pool, uint32_t value) nogil

    int dpiPool_setMaxSessionsPerShard(dpiPool *pool, uint32_t value) nogil

    int dpiPool_setPingInterval(dpiPool *pool, int value) nogil

    int dpiPool_setSodaMetadataCache(dpiPool *pool, bint enabled) nogil

    int dpiPool_setStmtCacheSize(dpiPool *pool, uint32_t cacheSize) nogil

    int dpiPool_setTimeout(dpiPool *pool, uint32_t value) nogil

    int dpiPool_setWaitTimeout(dpiPool *pool, uint32_t value) nogil

    int dpiQueue_deqMany(dpiQueue *queue, uint32_t *numProps,
            dpiMsgProps **props) nogil

    int dpiQueue_deqOne(dpiQueue *queue, dpiMsgProps **props) nogil

    int dpiQueue_enqMany(dpiQueue *queue, uint32_t numProps,
            dpiMsgProps **props) nogil

    int dpiQueue_enqOne(dpiQueue *queue, dpiMsgProps *props) nogil

    int dpiQueue_getDeqOptions(dpiQueue *queue,
            dpiDeqOptions **options) nogil

    int dpiQueue_getEnqOptions(dpiQueue *queue,
            dpiEnqOptions **options) nogil

    int dpiQueue_release(dpiQueue *queue) nogil

    int dpiRowid_getStringValue(dpiRowid *rowid, const char **value,
            uint32_t *valueLength)

    int dpiSodaColl_createIndex(dpiSodaColl *coll,
            const char *indexSpec, uint32_t indexSpecLength,
            uint32_t flags) nogil

    int dpiSodaColl_drop(dpiSodaColl *coll, uint32_t flags,
            bint *isDropped) nogil

    int dpiSodaColl_dropIndex(dpiSodaColl *coll, const char *name,
            uint32_t nameLength, uint32_t flags, bint *isDropped) nogil

    int dpiSodaColl_find(dpiSodaColl *coll,
            const dpiSodaOperOptions *options, uint32_t flags,
            dpiSodaDocCursor **cursor) nogil

    int dpiSodaColl_findOne(dpiSodaColl *coll,
            const dpiSodaOperOptions *options, uint32_t flags,
            dpiSodaDoc **doc) nogil

    int dpiSodaColl_getDataGuide(dpiSodaColl *coll, uint32_t flags,
            dpiSodaDoc **doc) nogil

    int dpiSodaColl_getDocCount(dpiSodaColl *coll,
            const dpiSodaOperOptions *options, uint32_t flags,
            uint64_t *count) nogil

    int dpiSodaColl_getMetadata(dpiSodaColl *coll, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaColl_getName(dpiSodaColl *coll, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaColl_insertManyWithOptions(dpiSodaColl *coll, uint32_t numDocs,
            dpiSodaDoc **docs, dpiSodaOperOptions *options, uint32_t flags,
            dpiSodaDoc **insertedDocs) nogil

    int dpiSodaColl_insertOneWithOptions(dpiSodaColl *coll, dpiSodaDoc *doc,
            dpiSodaOperOptions *options, uint32_t flags,
            dpiSodaDoc **insertedDoc) nogil

    int dpiSodaColl_release(dpiSodaColl *coll) nogil

    int dpiSodaColl_remove(dpiSodaColl *coll,
            const dpiSodaOperOptions *options, uint32_t flags,
            uint64_t *count) nogil

    int dpiSodaColl_replaceOne(dpiSodaColl *coll,
            const dpiSodaOperOptions *options, dpiSodaDoc *doc, uint32_t flags,
            bint *replaced, dpiSodaDoc **replacedDoc) nogil

    int dpiSodaColl_saveWithOptions(dpiSodaColl *coll, dpiSodaDoc *doc,
            dpiSodaOperOptions *options, uint32_t flags,
            dpiSodaDoc **savedDoc) nogil

    int dpiSodaColl_truncate(dpiSodaColl *coll) nogil

    int dpiSodaDb_createCollection(dpiSodaDb *db, const char *name,
            uint32_t nameLength, const char *metadata, uint32_t metadataLength,
            uint32_t flags, dpiSodaColl **coll) nogil

    int dpiSodaDb_createDocument(dpiSodaDb *db, const char *key,
            uint32_t keyLength, const char *content, uint32_t contentLength,
            const char *mediaType, uint32_t mediaTypeLength, uint32_t flags,
            dpiSodaDoc **doc) nogil

    int dpiSodaDb_freeCollectionNames(dpiSodaDb *db,
            dpiSodaCollNames *names) nogil

    int dpiSodaDb_getCollectionNames(dpiSodaDb *db,
            const char *startName, uint32_t startNameLength, uint32_t limit,
            uint32_t flags, dpiSodaCollNames *names) nogil

    int dpiSodaDb_openCollection(dpiSodaDb *db, const char *name,
            uint32_t nameLength, uint32_t flags, dpiSodaColl **coll) nogil

    int dpiSodaDb_release(dpiSodaDb *db) nogil

    int dpiSodaDoc_getContent(dpiSodaDoc *doc, const char **value,
            uint32_t *valueLength, const char **encoding) nogil

    int dpiSodaDoc_getCreatedOn(dpiSodaDoc *doc, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaDoc_getKey(dpiSodaDoc *doc, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaDoc_getLastModified(dpiSodaDoc *doc, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaDoc_getMediaType(dpiSodaDoc *doc, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaDoc_getVersion(dpiSodaDoc *doc, const char **value,
            uint32_t *valueLength) nogil

    int dpiSodaDoc_release(dpiSodaDoc *doc) nogil

    int dpiSodaDocCursor_close(dpiSodaDocCursor *cursor) nogil

    int dpiSodaDocCursor_getNext(dpiSodaDocCursor *cursor,
            uint32_t flags, dpiSodaDoc **doc) nogil

    int dpiSodaDocCursor_release(dpiSodaDocCursor *cursor) nogil

    int dpiStmt_addRef(dpiStmt *stmt) nogil

    int dpiStmt_bindByName(dpiStmt *stmt, const char *name,
            uint32_t nameLength, dpiVar *var) nogil

    int dpiStmt_bindByPos(dpiStmt *stmt, uint32_t pos, dpiVar *var) nogil

    int dpiStmt_close(dpiStmt *stmt, const char *tag, uint32_t tagLength)

    int dpiStmt_define(dpiStmt *stmt, uint32_t pos, dpiVar *var) nogil

    int dpiStmt_deleteFromCache(dpiStmt *stmt) nogil

    int dpiStmt_execute(dpiStmt *stmt, uint32_t mode,
            uint32_t *numQueryColumns) nogil

    int dpiStmt_executeMany(dpiStmt *stmt, uint32_t mode,
            uint32_t numIters) nogil

    int dpiStmt_fetchRows(dpiStmt *stmt, uint32_t maxRows,
            uint32_t *bufferRowIndex, uint32_t *numRowsFetched,
            bint *moreRows) nogil

    int dpiStmt_getBatchErrorCount(dpiStmt *stmt, uint32_t *count) nogil

    int dpiStmt_getBatchErrors(dpiStmt *stmt, uint32_t numErrors,
            dpiErrorInfo *errors) nogil

    int dpiStmt_getBindCount(dpiStmt *stmt, uint32_t *count) nogil

    int dpiStmt_getBindNames(dpiStmt *stmt, uint32_t *numBindNames,
            const char **bindNames, uint32_t *bindNameLengths) nogil

    int dpiStmt_getImplicitResult(dpiStmt *stmt,
            dpiStmt **implicitResult) nogil

    int dpiStmt_getNumQueryColumns(dpiStmt *stmt,
            uint32_t *numQueryColumns) nogil

    int dpiStmt_getInfo(dpiStmt *stmt, dpiStmtInfo *info) nogil

    int dpiStmt_getLastRowid(dpiStmt *stmt, dpiRowid **rowid) nogil

    int dpiStmt_getOciAttr(dpiStmt *stmt, uint32_t attribute,
            dpiDataBuffer *value, uint32_t *valueLength) nogil

    int dpiStmt_getQueryInfo(dpiStmt *stmt, uint32_t pos,
            dpiQueryInfo *info) nogil

    int dpiStmt_getRowCount(dpiStmt *stmt, uint64_t *count) nogil

    int dpiStmt_getRowCounts(dpiStmt *stmt, uint32_t *numRowCounts,
            uint64_t **rowCounts) nogil

    int dpiStmt_getSubscrQueryId(dpiStmt *stmt, uint64_t *queryId) nogil

    int dpiStmt_release(dpiStmt *stmt) nogil

    int dpiStmt_scroll(dpiStmt *stmt, uint32_t mode, int32_t offset,
            int32_t rowCountOffset) nogil

    int dpiStmt_setFetchArraySize(dpiStmt *stmt, uint32_t arraySize) nogil

    int dpiStmt_setOciAttr(dpiStmt *stmt, uint32_t attribute, void *value,
            uint32_t valueLength) nogil

    int dpiStmt_setPrefetchRows(dpiStmt *stmt, uint32_t numRows) nogil

    int dpiSubscr_prepareStmt(dpiSubscr *subscr, const char *sql,
            uint32_t sqlLength, dpiStmt **stmt) nogil

    int dpiSubscr_release(dpiSubscr *subscr) nogil

    int dpiVar_getNumElementsInArray(dpiVar *var, uint32_t *numElements) nogil

    int dpiVar_getReturnedData(dpiVar *var, uint32_t pos,
            uint32_t *numElements, dpiData **data) nogil

    int dpiVar_getSizeInBytes(dpiVar *var, uint32_t *sizeInBytes) nogil

    int dpiVar_release(dpiVar *var) nogil

    int dpiVar_setNumElementsInArray(dpiVar *var, uint32_t numElements) nogil

    int dpiVar_setFromBytes(dpiVar *var, uint32_t pos, const char *value,
            uint32_t valueLength) nogil

    int dpiVar_setFromLob(dpiVar *var, uint32_t pos, dpiLob *lob) nogil

    int dpiVar_setFromObject(dpiVar *var, uint32_t pos, dpiObject *obj) nogil

    int dpiVar_setFromStmt(dpiVar *var, uint32_t pos, dpiStmt *stmt) nogil



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# connection.pyx
#
# Cython file defining the thin Connection implementation class (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThinConnImpl(BaseConnImpl):

    cdef:
        object _statement_cache
        uint32_t _statement_cache_size
        object _statement_cache_lock
        Protocol _protocol
        str _server_version
        uint32_t _session_id
        uint32_t _serial_num
        str _action
        bint _action_modified
        str _dbop
        bint _dbop_modified
        str _client_info
        bint _client_info_modified
        str _client_identifier
        bint _client_identifier_modified
        str _module
        bint _module_modified
        ThinPoolImpl _pool
        bytes _ltxid
        str _current_schema
        bint _current_schema_modified
        str _edition
        str _internal_name
        str _external_name
        array.array _cursors_to_close
        ssize_t _num_cursors_to_close
        bint _drcp_enabled
        bint _drcp_establish_session
        double _time_in_pool
        list _temp_lobs_to_close
        uint32_t _temp_lobs_total_size
        uint32_t _call_timeout
        str _cclass
        int _dbobject_type_cache_num
        bytes _combo_key

    def __init__(self, str dsn, ConnectParamsImpl params):
        if not HAS_CRYPTOGRAPHY:
            errors._raise_err(errors.ERR_NO_CRYPTOGRAPHY_PACKAGE)
        BaseConnImpl.__init__(self, dsn, params)
        self._protocol = Protocol()

    cdef int _add_cursor_to_close(self, Statement stmt) except -1:
        if self._num_cursors_to_close == TNS_MAX_CURSORS_TO_CLOSE:
            raise Exception("too many cursors to close!")
        self._cursors_to_close[self._num_cursors_to_close] = stmt._cursor_id
        self._num_cursors_to_close += 1

    cdef int _adjust_statement_cache(self) except -1:
        cdef Statement stmt
        while len(self._statement_cache) > self._statement_cache_size:
            stmt = <Statement> self._statement_cache.popitem(last=False)[1]
            if stmt._in_use:
                stmt._return_to_cache = False
            elif stmt._cursor_id != 0:
                self._add_cursor_to_close(stmt)

    cdef int _connect_with_address(self, Address address,
                                   Description description,
                                   ConnectParamsImpl params,
                                   str connect_string,
                                   bint raise_exception) except -1:
        """
        Internal method used for connecting with the given description and
        address.
        """
        try:
            self._protocol._connect_phase_one(self, params, description,
                                              address, connect_string)
        except (exceptions.DatabaseError, socket.gaierror,
                ConnectionRefusedError) as e:
            if raise_exception:
                errors._raise_err(errors.ERR_CONNECTION_FAILED, cause=e)
            return 0
        except Exception as e:
            errors._raise_err(errors.ERR_CONNECTION_FAILED, cause=e)
        self._drcp_enabled = description.server_type == "pooled"
        if self._cclass is None:
            self._cclass = description.cclass
        self._protocol._connect_phase_two(self, description, params)

    cdef int _connect_with_description(self, Description description,
                                       ConnectParamsImpl params,
                                       bint final_desc) except -1:
        """
        Internal method used for connecting with the given description.
        """
        cdef:
            bint load_balance = description.load_balance
            bint raise_exc = False
            list address_lists = description.address_lists
            uint32_t i, j, k, num_addresses, idx1, idx2
            uint32_t num_attempts = description.retry_count + 1
            uint32_t num_lists = len(address_lists)
            AddressList address_list
            str connect_string
            Address address

        # Retry connecting to the socket if an attempt fails and retry_count
        # is specified in the connect string. If an attempt succeeds, return
        # the socket and the valid address object.
        connect_string = _get_connect_data(description)
        for i in range(num_attempts):

            # iterate through each address_list in the description; if the
            # description level load_balance is on, keep track of the least
            # recently used address for subsequent connections; otherwise,
            # iterate through the list in order; note that if source_route is
            # enabled that only the first address list is examined as the rest
            # are used for routing on the server
            if description.source_route and num_lists > 0:
                num_lists = 1
            for j in range(num_lists):
                if load_balance:
                    idx1 = (j + description.lru_index) % num_lists
                else:
                    idx1 = j

                # iterate through each address in an address_list; if the
                # address_list level load_balance is on, keep track of the
                # least recently used address for subsequent connections;
                # otherwise, iterate through the list in order; note that if
                # source_route is enabled that only the first address is
                # examined and the rest are used for routing on the server
                address_list = address_lists[idx1]
                num_addresses = len(address_list.addresses)
                if address_list.source_route and num_addresses > 0:
                    num_addresses = 1
                for k in range(num_addresses):
                    if address_list.load_balance:
                        idx2 = (k + address_list.lru_index) % num_addresses
                    else:
                        idx2 = k
                    address = address_list.addresses[idx2]
                    if final_desc:
                        raise_exc = i == num_attempts - 1 \
                                and j == num_lists - 1 \
                                and k == num_addresses - 1
                    self._connect_with_address(address, description, params,
                                               connect_string, raise_exc)
                    if self._protocol._in_connect:
                        continue
                    address_list.lru_index = (idx1 + 1) % num_addresses
                    description.lru_index = (idx2 + 1) % num_lists
                    return 0
            time.sleep(description.retry_delay)

    cdef int _connect_with_params(self, ConnectParamsImpl params) except -1:
        """
        Internal method used for connecting with the given parameters.
        """
        cdef:
            DescriptionList description_list = params.description_list
            list descriptions = description_list.descriptions
            ssize_t i, idx, num_descriptions = len(descriptions)
            Description description
            bint final_desc = False
        if description_list.source_route and num_descriptions > 0:
            num_descriptions = 1
        for i in range(num_descriptions):
            if i == num_descriptions - 1:
                final_desc = True
            if description_list.load_balance:
                idx = (i + description_list.lru_index) % num_descriptions
            else:
                idx = i
            description = descriptions[idx]
            self._connect_with_description(description, params, final_desc)
            if not self._protocol._in_connect:
                description_list.lru_index = (idx + 1) % num_descriptions
                break

    cdef Message _create_message(self, type typ):
        """
        Creates a message object that is used to send a request to the database
        and receive back its response.
        """
        cdef Message message
        message = typ.__new__(typ)
        message._initialize(self)
        return message

    cdef int _force_close(self) except -1:
        self._pool = None
        if self._dbobject_type_cache_num > 0:
            remove_dbobject_type_cache(self._dbobject_type_cache_num)
            self._dbobject_type_cache_num = 0
        self._protocol._force_close()

    cdef Statement _get_statement(self, str sql, bint cache_statement):
        """
        Get a statement from the statement cache, or prepare a new statement
        for use. If a statement is already in use a copy will be made and
        returned (and will not be returned to the cache). If a statement is
        being executed for the first time after releasing a DRCP session, a
        copy will also be made (and will not be returned to the cache) since it
        is unknown at this point whether the original session or a new session
        is going to be used.
        """
        cdef Statement statement
        with self._statement_cache_lock:
            statement = self._statement_cache.get(sql)
            if statement is None:
                statement = Statement()
                statement._prepare(sql)
                if cache_statement and not self._drcp_establish_session \
                        and not statement._is_ddl \
                        and self._statement_cache_size > 0:
                    statement._return_to_cache = True
                    self._statement_cache[sql] = statement
                    self._adjust_statement_cache()
            elif statement._in_use or not cache_statement \
                    or self._drcp_establish_session:
                if not cache_statement:
                    del self._statement_cache[sql]
                    statement._return_to_cache = False
                if statement._in_use or self._drcp_establish_session:
                    statement = statement.copy()
            else:
                self._statement_cache.move_to_end(sql)
            statement._in_use = True
            return statement

    cdef int _reset_statement_cache(self) except -1:
        """
        Reset the statement cache. This clears all statements and the list of
        cursors that need to be cleared.
        """
        with self._statement_cache_lock:
            self._statement_cache.clear()
            self._num_cursors_to_close = 0

    cdef int _return_statement(self, Statement statement) except -1:
        """
        Return the statement to the statement cache, if applicable. If the
        statement must not be returned to the statement cache, add the cursor
        id to the list of cursor ids to close on the next round trip to the
        database. Clear all bind variables and fetch variables in order to
        ensure that unnecessary references are not retained.
        """
        cdef:
            ThinVarImpl var_impl
            BindInfo bind_info
        if statement._bind_info_list is not None:
            for bind_info in statement._bind_info_list:
                bind_info._bind_var_impl = None
        if statement._fetch_var_impls is not None:
            for var_impl in statement._fetch_var_impls:
                var_impl._values = [None] * var_impl.num_elements
        with self._statement_cache_lock:
            if statement._return_to_cache:
                statement._in_use = False
            elif statement._cursor_id != 0:
                self._add_cursor_to_close(statement)

    def cancel(self):
        self._protocol._break_external()

    def change_password(self, str old_password, str new_password):
        cdef ChangePasswordMessage message
        message = self._create_message(ChangePasswordMessage)
        message.password = old_password.encode()
        message.newpassword = new_password.encode()
        self._protocol._process_single_message(message)

    def close(self, bint in_del=False):
        try:
            self._protocol._close(self)
        except (ssl.SSLError, exceptions.DatabaseError):
            pass

    def commit(self):
        cdef Message message
        message = self._create_message(CommitMessage)
        self._protocol._process_single_message(message)

    def connect(self, ConnectParamsImpl params):
        params._check_credentials()
        self._connect_with_params(params)
        self._statement_cache = collections.OrderedDict()
        self._statement_cache_size = params.stmtcachesize
        self._statement_cache_lock = threading.Lock()
        self._dbobject_type_cache_num = create_new_dbobject_type_cache(self)
        self._cursors_to_close = array.array('I')
        array.resize(self._cursors_to_close, TNS_MAX_CURSORS_TO_CLOSE)
        self.invoke_session_callback = True

    def create_cursor_impl(self):
        return ThinCursorImpl.__new__(ThinCursorImpl, self)

    def create_temp_lob_impl(self, DbType dbtype):
        return ThinLobImpl._create(self, dbtype)

    def get_call_timeout(self):
        return self._call_timeout

    def get_current_schema(self):
        return self._current_schema

    def get_edition(self):
        return self._edition

    def get_external_name(self):
        return self._external_name

    def get_internal_name(self):
        return self._internal_name

    def get_is_healthy(self):
        return self._protocol._socket is not None \
                and not self._protocol._read_buf._session_needs_to_be_closed

    def get_ltxid(self):
        return self._ltxid or b''

    def get_stmt_cache_size(self):
        return self._statement_cache_size

    def get_type(self, object conn, str name):
        cdef ThinDbObjectTypeCache cache = \
                get_dbobject_type_cache(self._dbobject_type_cache_num)
        return cache.get_type(conn, name)

    def get_version(self):
        return self._server_version

    def ping(self):
        cdef Message message
        message = self._create_message(PingMessage)
        self._protocol._process_single_message(message)

    def rollback(self):
        cdef Message message
        message = self._create_message(RollbackMessage)
        self._protocol._process_single_message(message)

    def set_action(self, str value):
        self._action = value
        self._action_modified = True

    def set_call_timeout(self, uint32_t value):
        timeout = None if value == 0 else value / 1000
        self._protocol._socket.settimeout(timeout)
        self._call_timeout = value

    def set_client_identifier(self, str value):
        self._client_identifier = value
        self._client_identifier_modified = True

    def set_client_info(self, str value):
        self._client_info = value
        self._client_info_modified = True

    def set_current_schema(self, value):
        self._current_schema = value
        self._current_schema_modified = True

    def set_dbop(self, str value):
        self._dbop = value
        self._dbop_modified = True

    def set_external_name(self, value):
        self._external_name = value

    def set_internal_name(self, value):
        self._internal_name = value

    def set_module(self, str value):
        self._module = value
        self._module_modified = True
        # setting the module by itself results in an error so always force the
        # action to be set as well (which eliminates this error)
        self._action_modified = True

    def set_stmt_cache_size(self, uint32_t value):
        self._statement_cache_size = value
        self._adjust_statement_cache()



#------------------------------------------------------------------------------
# Copyright (c) 2022, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# json.pyx
#
# Cython file defining the classes and methods used for encoding and decoding
# OSON (Oracle's extensions to JSON) (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class OsonDecoder(Buffer):

    cdef:
        ssize_t field_id_length
        ssize_t tree_seg_pos
        list field_names
        uint16_t flags

    cdef object _decode_container_node(self, uint8_t node_type):
        """
        Parses a container node (object or array) and returns it.
        """
        cdef:
            bint is_shared, is_object = (node_type & 0x40) == 0
            ssize_t field_ids_pos = 0, offsets_pos = 0, pos
            uint32_t i, num_children = 0
            uint32_t offset, temp32
            uint16_t temp16
            uint8_t temp8
            object value
            str name

        # determine the number of children by examining the 4th and 5th most
        # significant bits of the node type; determine the offsets in the tree
        # segment to the field ids array and the value offsets array
        self._get_num_children(node_type, &num_children, &is_shared)
        if is_shared:
            value = {}
            self._get_offset(node_type, &offset)
            offsets_pos = self._pos
            self.skip_to(self.tree_seg_pos + offset)
            self.read_ub1(&temp8)
            self._get_num_children(temp8, &num_children, &is_shared)
            field_ids_pos = self._pos
        elif is_object:
            value = {}
            field_ids_pos = self._pos
            offsets_pos = self._pos + self.field_id_length * num_children
        else:
            value = [None] * num_children
            offsets_pos = self._pos

        # process each of the children
        for i in range(num_children):
            if is_object:
                self.skip_to(field_ids_pos)
                if self.field_id_length == 1:
                    self.read_ub1(&temp8)
                    name = self.field_names[temp8 - 1]
                elif self.field_id_length == 2:
                    self.read_uint16(&temp16)
                    name = self.field_names[temp16 - 1]
                else:
                    self.read_uint32(&temp32)
                    name = self.field_names[temp32 - 1]
                field_ids_pos = self._pos
            self.skip_to(offsets_pos)
            self._get_offset(node_type, &offset)
            offsets_pos = self._pos
            self.skip_to(self.tree_seg_pos + offset)
            if is_object:
                value[name] = self._decode_node()
            else:
                value[i] = self._decode_node()

        return value

    cdef object _decode_node(self):
        """
        Parses a node from the tree segment and returns the Python equivalent.
        """
        cdef:
            uint8_t node_type, temp8
            const char_type* ptr
            uint16_t temp16
            uint32_t temp32

        # if the most significant bit is set the node refers to a container
        self.read_ub1(&node_type)
        if node_type & 0x80:
            return self._decode_container_node(node_type)

        # handle simple scalars
        if node_type == TNS_JSON_TYPE_NULL:
            return None
        elif node_type == TNS_JSON_TYPE_TRUE:
            return True
        elif node_type == TNS_JSON_TYPE_FALSE:
            return False

        # handle fixed length scalars
        elif node_type in (TNS_JSON_TYPE_DATE, TNS_JSON_TYPE_TIMESTAMP7):
            ptr = self._get_raw(7)
            return self.parse_date(ptr, 7)
        elif node_type == TNS_JSON_TYPE_TIMESTAMP:
            ptr = self._get_raw(11)
            return self.parse_date(ptr, 11)
        elif node_type == TNS_JSON_TYPE_TIMESTAMP_TZ:
            ptr = self._get_raw(13)
            return self.parse_date(ptr, 13)
        elif node_type == TNS_JSON_TYPE_BINARY_FLOAT:
            ptr = self._get_raw(4)
            return self.parse_binary_float(ptr)
        elif node_type == TNS_JSON_TYPE_BINARY_DOUBLE:
            ptr = self._get_raw(8)
            return self.parse_binary_double(ptr)
        elif node_type == TNS_JSON_TYPE_INTERVAL_DS:
            ptr = self._get_raw(11)
            return self.parse_interval_ds(ptr)
        elif node_type == TNS_JSON_TYPE_INTERVAL_YM:
            errors._raise_err(errors.ERR_DB_TYPE_NOT_SUPPORTED,
                              name="DB_TYPE_INTERVAL_YM")

        # handle scalars with lengths stored outside the node itself
        elif node_type == TNS_JSON_TYPE_STRING_LENGTH_UINT8:
            self.read_ub1(&temp8)
            ptr = self._get_raw(temp8)
            return ptr[:temp8].decode()
        elif node_type == TNS_JSON_TYPE_STRING_LENGTH_UINT16:
            self.read_uint16(&temp16)
            ptr = self._get_raw(temp16)
            return ptr[:temp16].decode()
        elif node_type == TNS_JSON_TYPE_STRING_LENGTH_UINT32:
            self.read_uint32(&temp32)
            ptr = self._get_raw(temp32)
            return ptr[:temp32].decode()
        elif node_type == TNS_JSON_TYPE_NUMBER_LENGTH_UINT8:
            return self.read_oracle_number(NUM_TYPE_DECIMAL)
        elif node_type == TNS_JSON_TYPE_BINARY_LENGTH_UINT16:
            self.read_uint16(&temp16)
            ptr = self._get_raw(temp16)
            return ptr[:temp16]
        elif node_type == TNS_JSON_TYPE_BINARY_LENGTH_UINT32:
            self.read_uint32(&temp32)
            ptr = self._get_raw(temp32)
            return ptr[:temp32]

        # handle number/decimal with length stored inside the node itself
        if (node_type & 0xf0) in (0x20, 0x60):
            temp8 = node_type & 0x0f
            ptr = self._get_raw(temp8 + 1)
            return self.parse_oracle_number(ptr, temp8 + 1, NUM_TYPE_DECIMAL)

        # handle integer with length stored inside the node itself
        elif (node_type & 0xf0) in (0x40, 0x50):
            temp8 = node_type & 0x0f
            ptr = self._get_raw(temp8)
            return self.parse_oracle_number(ptr, temp8, NUM_TYPE_DECIMAL)

        # handle string with length stored inside the node itself
        elif (node_type & 0xe0) == 0:
            if node_type == 0:
                return ''
            ptr = self._get_raw(node_type)
            return ptr[:node_type].decode()

        errors._raise_err(errors.ERR_OSON_NODE_TYPE_NOT_SUPPORTED,
                          node_type=node_type)

    cdef int _get_num_children(self, uint8_t node_type, uint32_t* num_children,
                               bint* is_shared) except -1:
        """
        Return the number of children the container has. This is determined by
        examining the 4th and 5th signficant bits of the node type:

            00 - number of children is uint8_t
            01 - number of children is uint16_t
            10 - number of children is uint32_t
            11 - field ids are shared with another object whose offset follows

        In the latter case the flag is_shared is set and the number of children
        is read by the caller instead as it must examine the offset and then
        retain the location for later use.
        """
        cdef:
            uint8_t temp8, children_bits = (node_type & 0x18)
            uint16_t temp16
        is_shared[0] = (children_bits == 0x18)
        if children_bits == 0:
            self.read_ub1(&temp8)
            num_children[0] = temp8
        elif children_bits == 0x08:
            self.read_uint16(&temp16)
            num_children[0] = temp16
        elif children_bits == 0x10:
            self.read_uint32(num_children)

    cdef int _get_offset(self, uint8_t node_type, uint32_t* offset) except -1:
        """
        Return an offset. The offset will be either a 16-bit or 32-bit value
        depending on the value of the 3rd significant bit of the node type.
        """
        cdef uint16_t temp16
        if node_type & 0x20:
            self.read_uint32(offset)
        else:
            self.read_uint16(&temp16)
            offset[0] = temp16

    cdef object decode(self, bytes data):
        """
        Returns a Python object corresponding to the encoded OSON bytes.
        """
        cdef:
            uint32_t num_field_names, field_names_seg_size, tree_seg_size, i
            ssize_t hash_id_size, field_name_offsets_size
            uint16_t num_tiny_nodes, temp16
            ssize_t field_name_offsets_pos
            uint8_t version, temp8
            const char_type* ptr
            uint32_t offset

        # populate the buffer with the data
        self._populate_from_bytes(data)

        # parse root header
        ptr = self._get_raw(3)
        if ptr[0] != TNS_JSON_MAGIC_BYTE_1 or \
                ptr[1] != TNS_JSON_MAGIC_BYTE_2 or \
                ptr[2] != TNS_JSON_MAGIC_BYTE_3:
            errors._raise_err(errors.ERR_UNEXPECTED_DATA, data=ptr[:3])
        self.read_ub1(&version)
        if version != TNS_JSON_VERSION:
            errors._raise_err(errors.ERR_OSON_VERSION_NOT_SUPPORTED,
                              version=version)
        self.read_uint16(&self.flags)

        # if value is a scalar value, the header is much smaller
        if self.flags & TNS_JSON_FLAG_IS_SCALAR:
            if self.flags & TNS_JSON_FLAG_TREE_SEG_UINT32:
                self.skip_raw_bytes(4)
            else:
                self.skip_raw_bytes(2)
            return self._decode_node()

        # determine the number of field names
        if self.flags & TNS_JSON_FLAG_NUM_FNAMES_UINT32:
            self.read_uint32(&num_field_names)
            self.field_id_length = 4
        elif self.flags & TNS_JSON_FLAG_NUM_FNAMES_UINT16:
            self.read_uint16(&temp16)
            num_field_names = temp16
            self.field_id_length = 2
        else:
            self.read_ub1(&temp8)
            num_field_names = temp8
            self.field_id_length = 1

        # determine the size of the field names segment
        if self.flags & TNS_JSON_FLAG_FNAMES_SEG_UINT32:
            field_name_offsets_size = 4
            self.read_uint32(&field_names_seg_size)
        else:
            field_name_offsets_size = 2
            self.read_uint16(&temp16)
            field_names_seg_size = temp16

        # determine the size of the tree segment
        if self.flags & TNS_JSON_FLAG_TREE_SEG_UINT32:
            self.read_uint32(&tree_seg_size)
        else:
            self.read_uint16(&temp16)
            tree_seg_size = temp16

        # determine the number of "tiny" nodes
        self.read_uint16(&num_tiny_nodes)

        # skip the hash id array
        if self.flags & TNS_JSON_FLAG_HASH_ID_UINT8:
            hash_id_size = 1
        elif self.flags & TNS_JSON_FLAG_HASH_ID_UINT16:
            hash_id_size = 2
        else:
            hash_id_size = 4
        self.skip_raw_bytes(num_field_names * hash_id_size)

        # skip the field name offsets array for now
        field_name_offsets_pos = self._pos
        self.skip_raw_bytes(num_field_names * field_name_offsets_size)
        ptr = self._get_raw(field_names_seg_size)

        # determine the names of the fields
        self.skip_to(field_name_offsets_pos)
        self.field_names = [None] * num_field_names
        for i in range(num_field_names):
            if self.flags & TNS_JSON_FLAG_FNAMES_SEG_UINT32:
                self.read_uint32(&offset)
            else:
                self.read_uint16(&temp16)
                offset = temp16
            temp8 = ptr[offset]
            self.field_names[i] = ptr[offset + 1:offset + temp8 + 1].decode()

        # get tree segment
        self.skip_raw_bytes(field_names_seg_size)
        self.tree_seg_pos = self._pos

        # return root node
        return self._decode_node()


@cython.final
cdef class OsonFieldName:

    cdef:
        str name
        bytes name_bytes
        ssize_t name_bytes_len
        uint32_t hash_id
        uint32_t offset
        uint32_t field_id

    cdef int _calc_hash_id(self) except -1:
        """
        Calculates the hash id to use for the field name. This is based on
        Bernstein's hash function.
        """
        cdef:
            const char_type *ptr = self.name_bytes
            ssize_t i
        self.hash_id = 0x811C9DC5
        for i in range(self.name_bytes_len):
            self.hash_id = (self.hash_id ^ ptr[i]) * 16777619

    @staticmethod
    cdef OsonFieldName create(str name):
        """
        Creates and initializes the field name.
        """
        cdef:
            OsonFieldName field_name
            ssize_t name_bytes_len
        field_name = OsonFieldName.__new__(OsonFieldName)
        field_name.name = name
        field_name.name_bytes = name.encode()
        name_bytes_len = len(field_name.name_bytes)
        if name_bytes_len > 255:
            errors._raise_err(errors.ERR_OSON_FIELD_NAME_LIMITATION)
        field_name.name_bytes_len = <uint8_t> name_bytes_len
        field_name._calc_hash_id()
        return field_name

    def sort_key(self):
        """
        Returns the sort key to use when sorting field names.
        """
        return ((self.hash_id & 0xff), self.name_bytes_len, self.name_bytes)


@cython.final
cdef class OsonFieldNamesSegment(GrowableBuffer):

    cdef:
        uint32_t num_field_names
        dict field_names_dict
        list field_names

    cdef int _examine_node(self, object value) except -1:
        """
        Examines the value. If it is a dictionary, all keys are extracted and
        unique names retained. Elements in lists and tuples and values in
        dictionaries are then examined to determine if they contain
        dictionaries as well.
        """
        cdef OsonFieldName field_name
        if isinstance(value, (list, tuple)):
            for child_value in value:
                self._examine_node(child_value)
        elif isinstance(value, dict):
            for key, child_value in (<dict> value).items():
                if key not in self.field_names_dict:
                    field_name = OsonFieldName.create(key)
                    self.field_names_dict[key] = field_name
                    field_name.offset = self._pos
                    self.write_uint8(field_name.name_bytes_len)
                    self.write_bytes(field_name.name_bytes)
                self._examine_node(child_value)

    cdef int _process_field_names(self) except -1:
        """
        Processes the field names in preparation for encoding within OSON.
        """
        cdef:
            OsonFieldName field_name
            ssize_t i
        self.field_names = sorted(self.field_names_dict.values(),
                                  key=OsonFieldName.sort_key)
        for i, field_name in enumerate(self.field_names):
            field_name.field_id = i + 1
        self.num_field_names = <uint32_t> len(self.field_names)

    @staticmethod
    cdef OsonFieldNamesSegment create(object value):
        """
        Creates and initializes the segment. The value (and all of its
        children) are examined for dictionaries and the keys retained as
        required by OSON.
        """
        cdef OsonFieldNamesSegment seg
        seg = OsonFieldNamesSegment.__new__(OsonFieldNamesSegment)
        seg._initialize(TNS_CHUNK_SIZE)
        seg.field_names_dict = {}
        seg._examine_node(value)
        seg._process_field_names()
        return seg


@cython.final
cdef class OsonTreeSegment(GrowableBuffer):

    cdef int _encode_container(self, uint8_t node_type,
                               ssize_t num_children) except -1:
        """
        Encodes the first part of a container.
        """
        node_type |= 0x20                   # use uint32_t for offsets
        if num_children > 65535:
            node_type |= 0x10               # num children is uint32_t
        elif num_children > 255:
            node_type |= 0x08               # num children is uint16_t
        self.write_uint8(node_type)
        if num_children < 256:
            self.write_uint8(<uint8_t> num_children)
        elif num_children < 65536:
            self.write_uint16(<uint16_t> num_children)
        else:
            self.write_uint32(<uint32_t> num_children)

    cdef int encode_array(self, object value,
                          OsonFieldNamesSegment fnames_seg) except -1:
        """
        Encode an array in the OSON tree segment.
        """
        cdef:
            ssize_t num_children
            uint8_t node_type
            uint32_t offset
        num_children = len(value)
        self._encode_container(TNS_JSON_TYPE_ARRAY, num_children)
        offset = self._pos
        self._reserve_space(num_children * sizeof(uint32_t))
        for element in value:
            pack_uint32(&self._data[offset], self._pos, BYTE_ORDER_MSB)
            offset += sizeof(uint32_t)
            self.encode_node(element, fnames_seg)

    cdef int encode_object(self, dict value,
                           OsonFieldNamesSegment fnames_seg) except -1:
        """
        Encode an object in the OSON tree segment.
        """
        cdef:
            uint32_t field_id_offset, value_offset
            uint8_t node_type, field_id_size
            OsonFieldName field_name
            ssize_t num_children
            object child_value
            str key
        num_children = len(value)
        self._encode_container(TNS_JSON_TYPE_OBJECT, num_children)
        if fnames_seg.num_field_names < 256:
            field_id_size = 1
        elif fnames_seg.num_field_names < 65536:
            field_id_size = 2
        else:
            field_id_size = 4
        field_id_offset = self._pos
        value_offset = self._pos + num_children * field_id_size
        self._reserve_space(num_children * (field_id_size + sizeof(uint32_t)))
        for key, child_value in value.items():
            field_name = fnames_seg.field_names_dict[key]
            if field_id_size == 1:
                self._data[field_id_offset] = <uint8_t> field_name.field_id
            elif field_id_size == 2:
                pack_uint16(&self._data[field_id_offset],
                            <uint16_t> field_name.field_id, BYTE_ORDER_MSB)
            else:
                pack_uint32(&self._data[field_id_offset], field_name.field_id,
                            BYTE_ORDER_MSB)
            pack_uint32(&self._data[value_offset], self._pos, BYTE_ORDER_MSB)
            field_id_offset += field_id_size
            value_offset += sizeof(uint32_t)
            self.encode_node(child_value, fnames_seg)

    cdef int encode_node(self, object value,
                         OsonFieldNamesSegment fnames_seg) except -1:
        """
        Encode a value (node) in the OSON tree segment.
        """
        cdef:
            uint32_t value_len
            bytes value_bytes

        # handle null
        if value is None:
            self.write_uint8(TNS_JSON_TYPE_NULL)

        # handle booleans
        elif isinstance(value, bool):
            if value is True:
                self.write_uint8(TNS_JSON_TYPE_TRUE)
            else:
                self.write_uint8(TNS_JSON_TYPE_FALSE)

        # handle numeric types
        elif isinstance(value, (int, float, PY_TYPE_DECIMAL)):
            value_bytes = (<str> cpython.PyObject_Str(value)).encode()
            self.write_uint8(TNS_JSON_TYPE_NUMBER_LENGTH_UINT8)
            self.write_oracle_number(value_bytes)

        # handle bytes
        elif isinstance(value, bytes):
            value_len = len(<bytes> value)
            if value_len < 65536:
                self.write_uint8(TNS_JSON_TYPE_BINARY_LENGTH_UINT16)
                self.write_uint16(<uint16_t> value_len)
            else:
                self.write_uint8(TNS_JSON_TYPE_BINARY_LENGTH_UINT32)
                self.write_uint32(value_len)
            self.write_bytes(<bytes> value)

        # handle timestamps
        elif isinstance(value, PY_TYPE_DATETIME):
            if cydatetime.PyDateTime_DATE_GET_MICROSECOND(value) == 0:
                self.write_uint8(TNS_JSON_TYPE_TIMESTAMP7)
                self.write_oracle_date(value, 7, write_length=False)
            else:
                self.write_uint8(TNS_JSON_TYPE_TIMESTAMP)
                self.write_oracle_date(value, 11, write_length=False)

        # handle dates
        elif isinstance(value, PY_TYPE_DATE):
            self.write_uint8(TNS_JSON_TYPE_DATE)
            self.write_oracle_date(value, 7, write_length=False)

        # handle timedeltas
        elif isinstance(value, PY_TYPE_TIMEDELTA):
            self.write_uint8(TNS_JSON_TYPE_INTERVAL_DS)
            self.write_interval_ds(value, write_length=False)

        # handle strings
        elif isinstance(value, str):
            value_bytes = (<str> value).encode()
            value_len = len(value_bytes)
            if value_len < 256:
                self.write_uint8(TNS_JSON_TYPE_STRING_LENGTH_UINT8)
                self.write_uint8(<uint8_t> value_len)
            elif value_len < 65536:
                self.write_uint8(TNS_JSON_TYPE_STRING_LENGTH_UINT16)
                self.write_uint16(<uint16_t> value_len)
            else:
                self.write_uint8(TNS_JSON_TYPE_STRING_LENGTH_UINT32)
                self.write_uint32(value_len)
            if value_len > 0:
                self.write_bytes(value_bytes)

        # handle lists/tuples
        elif isinstance(value, (list, tuple)):
            self.encode_array(value, fnames_seg)

        # handle dictionaries
        elif isinstance(value, dict):
            self.encode_object(value, fnames_seg)

        # other types are not supported
        else:
            errors._raise_err(errors.ERR_PYTHON_TYPE_NOT_SUPPORTED,
                              typ=type(value).__name__)


@cython.final
cdef class OsonEncoder(GrowableBuffer):

    cdef int encode(self, object value) except -1:
        """
        Encodes the given value to OSON.
        """
        cdef:
            OsonFieldNamesSegment fnames_seg = None
            OsonFieldName field_name
            OsonTreeSegment tree_seg
            uint16_t flags

        # determine the flags to use
        flags = TNS_JSON_FLAG_INLINE_LEAF
        if isinstance(value, (list, tuple, dict)):
            flags |= TNS_JSON_FLAG_HASH_ID_UINT8 | \
                     TNS_JSON_FLAG_TINY_NODES_STAT
            fnames_seg = OsonFieldNamesSegment.create(value);
            if fnames_seg.num_field_names > 65535:
                flags |= TNS_JSON_FLAG_NUM_FNAMES_UINT32
            elif fnames_seg.num_field_names > 255:
                flags |= TNS_JSON_FLAG_NUM_FNAMES_UINT16
            if fnames_seg._pos > 65535:
                flags |= TNS_JSON_FLAG_FNAMES_SEG_UINT32
        else:
            flags |= TNS_JSON_FLAG_IS_SCALAR

        # encode values into tree segment
        tree_seg = OsonTreeSegment.__new__(OsonTreeSegment)
        tree_seg._initialize(TNS_CHUNK_SIZE)
        tree_seg.encode_node(value, fnames_seg)
        if tree_seg._pos > 65535:
            flags |= TNS_JSON_FLAG_TREE_SEG_UINT32

        # write initial header
        self.write_uint8(TNS_JSON_MAGIC_BYTE_1)
        self.write_uint8(TNS_JSON_MAGIC_BYTE_2)
        self.write_uint8(TNS_JSON_MAGIC_BYTE_3)
        self.write_uint8(TNS_JSON_VERSION)
        self.write_uint16(flags)

        # write extended header (when value is not scalar)
        if fnames_seg is not None:

            # write number of field names
            if fnames_seg.num_field_names < 256:
                self.write_uint8(<uint8_t> fnames_seg.num_field_names)
            elif fnames_seg.num_field_names < 65536:
                self.write_uint16(<uint16_t> fnames_seg.num_field_names)
            else:
                self.write_uint32(fnames_seg.num_field_names)

            # write size of field names segment
            if fnames_seg._pos < 65536:
                self.write_uint16(<uint16_t> fnames_seg._pos)
            else:
                self.write_uint32(fnames_seg._pos)

        # write size of tree segment
        if (tree_seg._pos < 65536):
            self.write_uint16(<uint16_t> tree_seg._pos)
        else:
            self.write_uint32(tree_seg._pos)

        # write remainder of header and any data (when value is not scalar)
        if fnames_seg is not None:

            # write number of "tiny" nodes (always zero)
            self.write_uint16(0)

            # write array of hash ids
            for field_name in fnames_seg.field_names:
                self.write_uint8(field_name.hash_id & 0xff)

            # write array of field name offsets
            for field_name in fnames_seg.field_names:
                if fnames_seg._pos < 65536:
                    self.write_uint16(<uint16_t> field_name.offset)
                else:
                    self.write_uint32(field_name.offset)

            # write field names
            if fnames_seg._pos > 0:
                self.write_raw(fnames_seg._data, fnames_seg._pos)

        # write tree segment data
        self.write_raw(tree_seg._data, tree_seg._pos)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
1500 - Module for testing comparisons with database types and API types,
including the synonyms retained for backwards compatibility. This module also
tests for pickling/unpickling of database types and API types.
"""

import pickle

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def __test_compare(self, db_type, api_type):
        self.assertEqual(db_type, db_type)
        self.assertEqual(db_type, api_type)
        self.assertEqual(api_type, db_type)
        self.assertNotEqual(db_type, 5)
        self.assertNotEqual(db_type, oracledb.DB_TYPE_OBJECT)

    def __test_pickle(self, typ):
        self.assertIs(typ, pickle.loads(pickle.dumps(typ)))

    def test_1500_DB_TYPE_BFILE(self):
        "1500 - test oracledb.DB_TYPE_BFILE comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_BFILE, oracledb.BFILE)
        self.__test_pickle(oracledb.DB_TYPE_BFILE)

    def test_1501_DB_TYPE_BINARY_DOUBLE(self):
        "1501 - test oracledb.DB_TYPE_BINARY_DOUBLE comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_BINARY_DOUBLE, oracledb.NUMBER)
        self.assertEqual(oracledb.DB_TYPE_BINARY_DOUBLE, oracledb.NATIVE_FLOAT)
        self.__test_pickle(oracledb.DB_TYPE_BINARY_DOUBLE)

    def test_1502_DB_TYPE_BINARY_FLOAT(self):
        "1502 - test oracledb.DB_TYPE_BINARY_FLOAT comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_BINARY_FLOAT, oracledb.NUMBER)
        self.__test_pickle(oracledb.DB_TYPE_BINARY_FLOAT)

    def test_1503_DB_TYPE_BINARY_INTEGER(self):
        "1503 - test oracledb.DB_TYPE_BINARY_INTEGER comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_BINARY_INTEGER, oracledb.NUMBER)
        self.assertEqual(oracledb.DB_TYPE_BINARY_INTEGER, oracledb.NATIVE_INT)
        self.__test_pickle(oracledb.DB_TYPE_BINARY_INTEGER)

    def test_1504_DB_TYPE_BLOB(self):
        "1504 - test oracledb.DB_TYPE_BLOB comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_BLOB, oracledb.BLOB)
        self.__test_pickle(oracledb.DB_TYPE_BLOB)

    def test_1505_DB_TYPE_BOOLEAN(self):
        "1505 - test oracledb.DB_TYPE_BOOLEAN comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_BOOLEAN, oracledb.BOOLEAN)
        self.__test_pickle(oracledb.DB_TYPE_BOOLEAN)

    def test_1506_DB_TYPE_CHAR(self):
        "1506 - test oracledb.DB_TYPE_CHAR comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_CHAR, oracledb.STRING)
        self.assertEqual(oracledb.DB_TYPE_CHAR, oracledb.FIXED_CHAR)
        self.__test_pickle(oracledb.DB_TYPE_CHAR)

    def test_1507_DB_TYPE_CLOB(self):
        "1507 - test oracledb.DB_TYPE_CLOB comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_CLOB, oracledb.CLOB)
        self.__test_pickle(oracledb.DB_TYPE_CLOB)

    def test_1508_DB_TYPE_CURSOR(self):
        "1508 - test oracledb.DB_TYPE_CURSOR comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_CURSOR, oracledb.CURSOR)
        self.__test_pickle(oracledb.DB_TYPE_CURSOR)

    def test_1509_DB_TYPE_DATE(self):
        "1509 - test oracledb.DB_TYPE_DATE comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_DATE, oracledb.DATETIME)
        self.__test_pickle(oracledb.DB_TYPE_DATE)

    def test_1510_DB_TYPE_INTERVAL_DS(self):
        "1510 - test oracledb.DB_TYPE_INTERVAL_DS comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_INTERVAL_DS, oracledb.INTERVAL)
        self.__test_pickle(oracledb.DB_TYPE_INTERVAL_DS)

    def test_1511_DB_TYPE_LONG(self):
        "1511 - test oracledb.DB_TYPE_LONG comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_LONG, oracledb.STRING)
        self.assertEqual(oracledb.DB_TYPE_LONG, oracledb.LONG_STRING)
        self.__test_pickle(oracledb.DB_TYPE_LONG)

    def test_1512_DB_TYPE_LONG_RAW(self):
        "1512 - test oracledb.DB_TYPE_LONG_RAW comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_LONG_RAW, oracledb.BINARY)
        self.assertEqual(oracledb.DB_TYPE_LONG_RAW, oracledb.LONG_BINARY)
        self.__test_pickle(oracledb.DB_TYPE_LONG_RAW)

    def test_1513_DB_TYPE_NCHAR(self):
        "1513 - test oracledb.DB_TYPE_NCHAR comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_NCHAR, oracledb.STRING)
        self.assertEqual(oracledb.DB_TYPE_NCHAR, oracledb.FIXED_NCHAR)
        self.__test_pickle(oracledb.DB_TYPE_NCHAR)

    def test_1514_DB_TYPE_NCLOB(self):
        "1514 - test oracledb.DB_TYPE_NCLOB comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_NCLOB, oracledb.NCLOB)
        self.__test_pickle(oracledb.DB_TYPE_NCLOB)

    def test_1515_DB_TYPE_NUMBER(self):
        "1515 - test oracledb.DB_TYPE_NUMBER comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_NUMBER, oracledb.NUMBER)
        self.__test_pickle(oracledb.DB_TYPE_NUMBER)

    def test_1516_DB_TYPE_NVARCHAR(self):
        "1516 - test oracledb.DB_TYPE_NVARCHAR comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_NVARCHAR, oracledb.STRING)
        self.assertEqual(oracledb.DB_TYPE_NVARCHAR, oracledb.NCHAR)
        self.__test_pickle(oracledb.DB_TYPE_NVARCHAR)

    def test_1517_DB_TYPE_OBJECT(self):
        "1517 - test oracledb.DB_TYPE_OBJECT comparisons and pickling"
        self.assertEqual(oracledb.DB_TYPE_OBJECT, oracledb.OBJECT)
        self.__test_pickle(oracledb.DB_TYPE_OBJECT)

    def test_1518_DB_TYPE_RAW(self):
        "1518 - test oracledb.DB_TYPE_RAW comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_RAW, oracledb.BINARY)
        self.__test_pickle(oracledb.DB_TYPE_RAW)

    def test_1519_DB_TYPE_ROWID(self):
        "1519 - test oracledb.DB_TYPE_ROWID comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_ROWID, oracledb.ROWID)
        self.__test_pickle(oracledb.DB_TYPE_ROWID)

    def test_1520_DB_TYPE_TIMESTAMP(self):
        "1520 - test oracledb.DB_TYPE_TIMESTAMP comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_TIMESTAMP, oracledb.DATETIME)
        self.assertEqual(oracledb.DB_TYPE_TIMESTAMP, oracledb.TIMESTAMP)
        self.__test_pickle(oracledb.DB_TYPE_TIMESTAMP)

    def test_1521_DB_TYPE_TIMESTAMP_LTZ(self):
        "1521 - test oracledb.DB_TYPE_TIMESTAMP_LTZ comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_TIMESTAMP_LTZ, oracledb.DATETIME)
        self.__test_pickle(oracledb.DB_TYPE_TIMESTAMP_LTZ)

    def test_1522_DB_TYPE_TIMESTAMP_TZ(self):
        "1522 - test oracledb.DB_TYPE_TIMESTAMP_TZ comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_TIMESTAMP_TZ, oracledb.DATETIME)
        self.__test_pickle(oracledb.DB_TYPE_TIMESTAMP_TZ)

    def test_1523_DB_TYPE_VARCHAR(self):
        "1523 - test oracledb.DB_TYPE_VARCHAR comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_VARCHAR, oracledb.STRING)
        self.__test_pickle(oracledb.DB_TYPE_VARCHAR)

    def test_1524_NUMBER(self):
        "1524 - test oracledb.NUMBER pickling"
        self.__test_pickle(oracledb.NUMBER)

    def test_1525_STRING(self):
        "1525 - test oracledb.STRING pickling"
        self.__test_pickle(oracledb.STRING)

    def test_1526_DATETIME(self):
        "1526 - test oracledb.DATETIME pickling"
        self.__test_pickle(oracledb.DATETIME)

    def test_1527_BINARY(self):
        "1527 - test oracledb.BINARY pickling"
        self.__test_pickle(oracledb.BINARY)

    def test_1528_ROWID(self):
        "1528 - test oracledb.ROWID pickling"
        self.__test_pickle(oracledb.ROWID)

    def test_1529_DB_TYPE_UROWID(self):
        "1529 - test oracledb.DB_TYPE_UROWID comparisons and pickling"
        self.__test_compare(oracledb.DB_TYPE_UROWID, oracledb.ROWID)
        self.__test_pickle(oracledb.DB_TYPE_UROWID)

    def test_1530_DB_TYPE_JSON(self):
        "1530 - test oracledb.DB_TYPE_JSON pickling"
        self.__test_pickle(oracledb.DB_TYPE_JSON)

    def test_1531_DB_TYPE_INTERVAL_YM(self):
        "1531 - test oracledb.DB_TYPE_INTERVAL_YM pickling"
        self.__test_pickle(oracledb.DB_TYPE_INTERVAL_YM)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
5000 - Module for testing external authentication
"""

import unittest

import oracledb
import test_env

@unittest.skipIf(not test_env.get_external_user(),
                 "external authentication not supported with this setup")
class TestCase(test_env.BaseTestCase):
    require_connection = False

    def __verify_connection(self, connection, expected_user,
                            expected_proxy_user=None):
        with connection.cursor() as cursor:
            cursor.execute("""
                    select
                        sys_context('userenv', 'session_user'),
                        sys_context('userenv', 'proxy_user')
                    from dual""")
            actual_user, actual_proxy_user = cursor.fetchone()
            self.assertEqual(actual_user, expected_user.upper())
            self.assertEqual(actual_proxy_user,
                             expected_proxy_user and expected_proxy_user.upper())

    def test_5000_pool_with_user_and_password_and_externalauth_enabled(self):
        """
        5000 - test error on creating a pool with user and password specified
        and externalauth enabled
        """
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1032:",
                               test_env.get_pool, min=1, max=2, increment=1,
                               getmode=oracledb.POOL_GETMODE_WAIT,
                               externalauth=True, homogeneous=False)

    def test_5001_pool_with_no_password_and_externalauth_enabled(self):
        """
        5001 - test error on creating a pool without password and with user
        specified and externalauth enabled
        """
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1032:",
                               oracledb.create_pool,
                               user=test_env.get_main_user(),
                               min=1, max=2, increment=1,
                               getmode=oracledb.POOL_GETMODE_WAIT,
                               externalauth=True, homogeneous=False)

    def test_5002_pool_with_no_user_and_externalauth_enabled(self):
        """
        5002 - test error on creating a pool without user and with password
        specified and externalauth enabled
        """
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1032:",
                               oracledb.create_pool,
                               password=test_env.get_main_password(),
                               min=1, max=2, increment=1,
                               getmode=oracledb.POOL_GETMODE_WAIT,
                               externalauth=True, homogeneous=False)

    def test_5003_pool_with_user_and_password_and_externalauth_off(self):
        """
        5003 - test creating a pool with user and password specified and
        externalauth set to False
        """
        pool = test_env.get_pool(min=1, max=2, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT,
                                 externalauth=False, homogeneous=False)
        with pool.acquire() as connection:
            self.__verify_connection(connection, test_env.get_main_user())

    def test_5004_user_and_password_with_externalauth_enabled(self):
        """
        5004 - test error when connecting with user and password specified
        and externalauth enabled
        """
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1032:",
                               oracledb.connect, user=test_env.get_main_user(),
                               password=test_env.get_main_password(),
                               dsn=test_env.get_connect_string(),
                               externalauth=True)

    def test_5005_no_user_with_externalauth_enabled(self):
        """
        5005 - test error when connecting without username and with password
        specified and externalauth enabled
        """
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1032:",
                               oracledb.connect,
                               password=test_env.get_main_password(),
                               dsn=test_env.get_connect_string(),
                               externalauth=True)

        # by default externalauth is False
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-01017:",
                               oracledb.connect,
                               password=test_env.get_main_password(),
                               dsn=test_env.get_connect_string())

    def test_5006_user_with_no_password_and_externalauth_enabled(self):
        """
        5006 - test error when connecting without password and with user
        specified and externalauth enabled
        """
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-01017:",
                               oracledb.connect, user="[invalid_user]",
                               dsn=test_env.get_connect_string(),
                               externalauth=True)

        # by default externalauth is False
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-01017:",
                               oracledb.connect, user="[invalid_user]",
                               dsn=test_env.get_connect_string())

    def test_5007_external_authentication_with_invalid_proxy_user(self):
        "5007 - test external authentication with invalid proxy user"
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1069:",
                               oracledb.connect, user=test_env.get_main_user(),
                               dsn=test_env.get_connect_string(),
                               externalauth=True)

        # by default externalauth is False
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4001:",
                               oracledb.connect, user=test_env.get_main_user(),
                               dsn=test_env.get_connect_string())

    def test_5008_user_and_password_with_externalauth_off(self):
        """
        5008 - test creating a connection with user and password specified and
        externalauth set to False
        """
        connection = oracledb.connect(user=test_env.get_main_user(),
                                      password=test_env.get_main_password(),
                                      dsn=test_env.get_connect_string(),
                                      externalauth=False)
        self.__verify_connection(connection, test_env.get_main_user())

    def test_5009_external_authentication_with_externalauth_enabled(self):
        """
        5009 - test creating standalone connection with externalauth set to
        True explicitly
        """
        connection = oracledb.connect(dsn=test_env.get_connect_string(),
                                      externalauth=True)
        self.__verify_connection(connection, test_env.get_external_user())

    def test_5010_external_authentication_with_externalauth_not_set(self):
        """
        5010 - test creating standalone connection with no user and password
        specified and externalauth not set
        """
        connection = oracledb.connect(dsn=test_env.get_connect_string())
        self.__verify_connection(connection, test_env.get_external_user())

    def test_5011_pool_with_external_authentication(self):
        "5011 - test creating a pool with external authentication"
        pool = oracledb.create_pool(dsn=test_env.get_connect_string(),
                                    min=1, max=2, increment=1,
                                    getmode=oracledb.POOL_GETMODE_WAIT,
                                    externalauth=True, homogeneous=False)
        self.assertEqual(pool.opened, 0)
        with pool.acquire() as connection:
            self.__verify_connection(connection, test_env.get_external_user())

    def test_5012_pool_with_no_user_and_password_and_externalauth_not_set(self):
        """
        5012 - test creating a pool without user and password specified and
        externalauth not set
        """
        pool = oracledb.create_pool(dsn=test_env.get_connect_string(),
                                    min=1, max=2, increment=1,
                                    getmode=oracledb.POOL_GETMODE_WAIT,
                                    homogeneous=False)
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-24415:",
                               pool.acquire)

    def test_5013_pool_min_with_no_effect_under_external_authentication(self):
        "5013 - test pool min is always 0 under external authentication"
        pool = oracledb.create_pool(dsn=test_env.get_connect_string(),
                                    min=5, max=10, increment=3,
                                    getmode=oracledb.POOL_GETMODE_WAIT,
                                    externalauth=True, homogeneous=False)
        self.assertEqual(pool.opened, 0)

    def test_5014_pool_increment_with_no_effect_under_external_auth(self):
        "5014 - test pool increment is always 1 under external authentication"
        pool = oracledb.create_pool(dsn=test_env.get_connect_string(),
                                    min=5, max=10, increment=3,
                                    getmode=oracledb.POOL_GETMODE_WAIT,
                                    externalauth=True, homogeneous=False)
        conn1 = pool.acquire()
        self.assertEqual(pool.opened, 1)
        conn2 = pool.acquire()
        self.assertEqual(pool.opened, 2)

    def test_5015_external_authentication_with_proxy(self):
        "5015 - test external authentication with proxy"
        proxy_user = test_env.get_external_user() # proxy user
        schema_user = test_env.get_main_user()    # schema user
        conn1 = oracledb.connect(user=f"[{schema_user}]",
                                 dsn=test_env.get_connect_string(),
                                 externalauth=True)
        self.__verify_connection(conn1, schema_user, proxy_user)
        conn2 = oracledb.connect(user=f"[{schema_user}]",
                                 dsn=test_env.get_connect_string())
        self.__verify_connection(conn2, schema_user, proxy_user)

    def test_5016_pool_external_authentication_with_proxy(self):
        "5016 - test creating pool using external authentication with proxy"
        proxy_user = test_env.get_external_user()
        schema_user = test_env.get_main_user()
        pool = oracledb.create_pool(externalauth=True, homogeneous=False,
                                    dsn=test_env.get_connect_string(),
                                    min=1, max=2, increment=1,
                                    getmode=oracledb.POOL_GETMODE_WAIT)
        self.assertEqual(pool.opened, 0)
        connection = pool.acquire(user=f"[{schema_user}]")
        self.__verify_connection(connection, schema_user, proxy_user)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
4400 - Module for testing TPC (two-phase commit) transactions.
"""

import unittest

import oracledb
import test_env

@unittest.skipIf(test_env.get_is_thin(),
                 "thin mode doesn't support two-phase commit yet")
class TestCase(test_env.BaseTestCase):

    def test_4400_tpc_with_rolback(self):
        "4400 - test begin, prepare, roll back global transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.connection.xid(3900, "txn3900", "branchId")
        self.connection.tpc_begin(xid)
        self.assertEqual(self.connection.tpc_prepare(), False)
        self.connection.tpc_begin(xid)
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'tesName')""")
        self.assertEqual(self.connection.tpc_prepare(), True)
        self.connection.tpc_rollback()
        self.cursor.execute("select count(*) from TestTempTable")
        count, = self.cursor.fetchone()
        self.assertEqual(count, 0)

    def test_4401_tpc_with_commit(self):
        "4401 - test begin, prepare, commit global transaction"
        self.cursor.execute("truncate table TestTempTable")
        xid = self.connection.xid(3901, "txn3901", "branchId")
        self.connection.tpc_begin(xid)
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'tesName')""")
        self.assertEqual(self.connection.tpc_prepare(), True)
        self.connection.tpc_commit()
        self.cursor.execute("select IntCol, StringCol1 from TestTempTable")
        self.assertEqual(self.cursor.fetchall(), [(1, 'tesName')])

    def test_4402_tpc_multiple_transactions(self):
        "4402 - test multiple global transactions on the same connection"
        self.cursor.execute("truncate table TestTempTable")
        xid1 = self.connection.xid(3902, "txn3902", "branch1")
        xid2 = self.connection.xid(3902, "txn3902", "branch2")
        self.connection.tpc_begin(xid1)
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (1, 'tesName')""")
        self.connection.tpc_end()
        self.connection.tpc_begin(xid2)
        self.cursor.execute("""
                insert into TestTempTable (IntCol, StringCol1)
                values (2, 'tesName')""")
        self.connection.tpc_end()
        needs_commit1 = self.connection.tpc_prepare(xid1)
        needs_commit2 = self.connection.tpc_prepare(xid2)
        if needs_commit1:
            self.connection.tpc_commit(xid1)
        if needs_commit2:
            self.connection.tpc_commit(xid2)
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                order by IntCol""")
        expected_rows = [(1, 'tesName'), (2, 'tesName')]
        self.assertEqual(self.cursor.fetchall(), expected_rows)

    def test_4403_rollback_with_xid(self):
        "4403 - test rollback with parameter xid"
        self.cursor.execute("truncate table TestTempTable")
        xid1 = self.connection.xid(3901, "txn3901", "branch1")
        xid2 = self.connection.xid(3902, "txn3902", "branch2")
        for count, xid in enumerate([xid1, xid2]):
            self.connection.tpc_begin(xid)
            self.cursor.execute("""
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:id, 'tesName')""", id=count)
            self.connection.tpc_end()
        self.connection.tpc_rollback(xid1)

        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-24756",
                               self.connection.tpc_prepare, xid1)
        needs_commit = self.connection.tpc_prepare(xid2)
        if needs_commit:
            self.connection.tpc_commit(xid2)
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                order by IntCol""")
        self.assertEqual(self.cursor.fetchall(), [(1, 'tesName')])

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# pool_params.pyx
#
# Cython file defining the base PoolParams implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class PoolParamsImpl(ConnectParamsImpl):

    def __init__(self):
        ConnectParamsImpl.__init__(self)
        self.min = 1
        self.max = 2
        self.increment = 1
        self.getmode = constants.POOL_GETMODE_WAIT
        self.homogeneous = True
        self.ping_interval = 60

    cdef int _copy(self, ConnectParamsImpl other_params) except -1:
        """
        Internal method for copying attributes from another set of parameters.
        """
        cdef PoolParamsImpl pool_params = <PoolParamsImpl> other_params
        ConnectParamsImpl._copy(self, other_params)
        self.min = pool_params.min
        self.max = pool_params.max
        self.increment = pool_params.increment
        self.connectiontype = pool_params.connectiontype
        self.getmode = pool_params.getmode
        self.homogeneous = pool_params.homogeneous
        self.timeout = pool_params.timeout
        self.wait_timeout = pool_params.wait_timeout
        self.max_lifetime_session = pool_params.max_lifetime_session
        self.session_callback = pool_params.session_callback
        self.max_sessions_per_shard = pool_params.max_sessions_per_shard
        self.soda_metadata_cache = pool_params.soda_metadata_cache
        self.ping_interval = pool_params.ping_interval

    def copy(self):
        """
        Creates a copy of the connection parameters and returns it.
        """
        cdef PoolParamsImpl new_params
        new_params = PoolParamsImpl.__new__(PoolParamsImpl)
        new_params._copy(self)
        return new_params

    def set(self, dict args):
        """
        Sets the property values based on the supplied arguments. All values
        not supplied will be left unchanged.
        """
        ConnectParamsImpl.set(self, args)
        _set_uint_param(args, "min", &self.min)
        _set_uint_param(args, "max", &self.max)
        _set_uint_param(args, "increment", &self.increment)
        self.connectiontype = args.get("connectiontype")
        _set_uint_param(args, "getmode", &self.getmode)
        _set_bool_param(args, "homogeneous", &self.homogeneous)
        _set_uint_param(args, "timeout", &self.timeout)
        _set_uint_param_with_deprecated_name(args, "wait_timeout",
                                             "waitTimeout", &self.wait_timeout)
        _set_uint_param_with_deprecated_name(args, "max_lifetime_session",
                                             "maxLifetimeSession",
                                             &self.max_lifetime_session)
        _set_obj_param_with_deprecated_name(args, "session_callback",
                                            "sessionCallback", self)
        _set_uint_param_with_deprecated_name(args, "max_sessions_per_shard",
                                             "maxSessionsPerShard",
                                             &self.max_sessions_per_shard)
        _set_bool_param(args, "soda_metadata_cache", &self.soda_metadata_cache)
        _set_int_param(args, "ping_interval", &self.ping_interval)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# create_schema.py
#
# Creates users and populates their schemas with the tables and packages
# necessary for running the python-oracledb test suite.
#------------------------------------------------------------------------------

import oracledb

import drop_schema
import test_env

# connect as administrative user (usually SYSTEM or ADMIN)
conn = test_env.get_admin_connection()

# drop existing users and editions, if applicable
drop_schema.drop_schema(conn)

# create test schemas
print("Creating test schemas...")
test_env.run_sql_script(conn, "create_schema",
                        main_user=test_env.get_main_user(),
                        main_password=test_env.get_main_password(),
                        proxy_user=test_env.get_proxy_user(),
                        proxy_password=test_env.get_proxy_password())
if test_env.get_server_version() >= (21, 0):
    test_env.run_sql_script(conn, "create_schema_21",
                            main_user=test_env.get_main_user())
if test_env.is_on_oracle_cloud(conn):
    test_env.run_sql_script(conn, "create_schema_cloud",
                            main_user=test_env.get_main_user())
print("Done.")



#------------------------------------------------------------------------------
# Copyright (c) 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
4700 - Module for testing pool parameters.
"""

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def __test_writable_parameter(self, name, value):
        """
        Tests that a writable parameter can be written to and the modified
        value read back successfully.
        """
        params = oracledb.PoolParams()
        orig_value = getattr(params, name)
        copied_params = params.copy()
        args = {}
        args[name] = value
        params.set(**args)
        self.assertEqual(getattr(params, name), value)
        self.assertEqual(getattr(copied_params, name), orig_value)

    def test_4700_writable_params(self):
        "4700 - test writable parameters"
        self.__test_writable_parameter("min", 8)
        self.__test_writable_parameter("max", 12)
        self.__test_writable_parameter("increment", 2)
        self.__test_writable_parameter("connectiontype", oracledb.Connection)
        self.__test_writable_parameter("getmode", oracledb.POOL_GETMODE_NOWAIT)
        self.__test_writable_parameter("homogeneous", False)
        self.__test_writable_parameter("timeout", 25)
        self.__test_writable_parameter("wait_timeout", 45)
        self.__test_writable_parameter("max_lifetime_session", 65)
        self.__test_writable_parameter("session_callback", lambda c: None)
        self.__test_writable_parameter("max_sessions_per_shard", 5)
        self.__test_writable_parameter("soda_metadata_cache", True)
        self.__test_writable_parameter("ping_interval", 20)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2900 - Module for testing Rowids
"""

import datetime
import unittest

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def __populate_test_universal_rowids(self):
        self.cursor.execute("truncate table TestUniversalRowids")
        data = [
            (1, "ABC" * 75, datetime.datetime(2017, 4, 11)),
            (2, "DEF" * 80, datetime.datetime(2017, 4, 12))
        ]
        self.cursor.executemany("""
                    insert into TestUniversalRowids
                    values (:1, :2, :3)""", data)
        self.connection.commit()

    def __test_select_rowids(self, table_name):
        self.cursor.execute(f"select rowid, IntCol from {table_name}")
        sql = f"select IntCol from {table_name} where rowid = :val"
        for rowid, int_val in self.cursor.fetchall():
            self.cursor.execute(sql, val=rowid)
            self.assertEqual(self.cursor.fetchall(), [(int_val,)])

    def test_2900_select_rowids_regular(self):
        "2900 - test selecting all rowids from a regular table"
        self.__test_select_rowids("TestNumbers")

    def test_2901_select_rowids_index_organised(self):
        "2901 - test selecting all rowids from an index organised table"
        self.__populate_test_universal_rowids()
        self.__test_select_rowids("TestUniversalRowids")

    def test_2902_insert_invalid_rowid(self):
        "2902 - test inserting an invalid rowid"
        sql = "insert into TestRowids (IntCol, RowidCol) values (1, :rid)"
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-00932:",
                               self.cursor.execute, sql, rid=12345)
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-01410:",
                               self.cursor.execute, sql, rid="523lkhlf")

    def test_2903_select_rowids_as_urowids(self):
        "2903 - test selecting regular rowids stored in a urowid column"
        self.cursor.execute("truncate table TestRowids")
        self.cursor.execute("""
                insert into TestRowids (IntCol, UrowidCol)
                select IntCol, rowid from TestNumbers""")
        self.connection.commit()
        self.cursor.execute("select IntCol, UrowidCol from TestRowids")
        for int_val, rowid in self.cursor.fetchall():
            self.cursor.execute("""
                    select IntCol
                    from TestNumbers
                    where rowid = :val""",
                    val=rowid)
            self.assertEqual(self.cursor.fetchall(), [(int_val,)])

    def test_2904_select_rowids_as_rowids(self):
        "2904 - test selecting regular rowids stored in a rowid column"
        self.cursor.execute("truncate table TestRowids")
        self.cursor.execute("""
                insert into TestRowids (IntCol, RowidCol)
                select IntCol, rowid from TestNumbers""")
        self.connection.commit()
        self.cursor.execute("select IntCol, RowidCol from TestRowids")
        for int_val, rowid in self.cursor.fetchall():
            self.cursor.execute("""
                    select IntCol
                    from TestNumbers
                    where rowid = :val""",
                    val=rowid)
            self.assertEqual(self.cursor.fetchall(), [(int_val,)])

    def test_2905_test_bind_and_insert_rowid(self):
        "2905 - binding and inserting a rowid"
        self.cursor.execute("truncate table TestRowids")
        insert_data = [
            (1, "String #1"),
            (2, "String #2"),
            (3, "String #3"),
            (4, "String #4")
        ]
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.executemany(sql, insert_data)
        self.connection.commit()
        ridvar = self.cursor.var(oracledb.ROWID)
        self.cursor.execute("""
            begin
                select rowid into :rid from TestTempTable
                where IntCol = 3;
            end;""",
            rid=ridvar)
        self.cursor.setinputsizes(r1=oracledb.ROWID)
        self.cursor.execute("""
            insert into TestRowids (IntCol, RowidCol)
            values(1, :r1)""",
            r1=ridvar)
        self.connection.commit()
        self.cursor.execute("select IntCol, RowidCol from TestRowids")
        int_val, rowid = self.cursor.fetchone()
        self.cursor.execute("""
            select IntCol, StringCol1 from TestTempTable
            where rowid = :val""",
            val=rowid)
        self.assertEqual(self.cursor.fetchone(), (3, "String #3"))

    @unittest.skipIf(not test_env.get_is_thin(),
                     "thick mode doesn't support DB_TYPE_UROWID")
    def test_2906_test_bind_and_insert_rowid_as_urowid(self):
        "2906 - binding and inserting a rowid as urowid"
        self.cursor.execute("truncate table TestRowids")
        insert_data = [
            (1, "String #1", datetime.datetime(2017, 4, 4)),
            (2, "String #2", datetime.datetime(2017, 4, 5)),
            (3, "String #3", datetime.datetime(2017, 4, 6)),
            (4, "A" * 250, datetime.datetime(2017, 4, 7))
        ]
        self.cursor.execute("truncate table TestUniversalRowids")
        sql = "insert into TestUniversalRowids values (:1, :2, :3)"
        self.cursor.executemany(sql, insert_data)
        self.connection.commit()
        ridvar = self.cursor.var(oracledb.DB_TYPE_UROWID)
        self.cursor.execute("""
            begin
                select rowid into :rid from TestUniversalRowids
                where IntCol = 3;
            end;""",
            rid=ridvar)
        self.cursor.setinputsizes(r1=oracledb.DB_TYPE_UROWID)
        self.cursor.execute("""
            insert into TestRowids (IntCol, UrowidCol)
            values(1, :r1)""",
            r1=ridvar)
        self.connection.commit()
        self.cursor.execute("select IntCol, UrowidCol from TestRowids")
        int_val, rowid = self.cursor.fetchone()
        self.cursor.execute("""
            select IntCol, StringCol, DateCol from TestUniversalRowids
            where rowid = :val""",
            val=rowid)
        self.assertEqual(self.cursor.fetchone(),
                         (3, "String #3", datetime.datetime(2017, 4, 6)))

    def test_2907_test_fetch_null_rowids(self):
        "2907 - fetching a null rowid"
        self.cursor.execute("truncate table TestRowids")
        self.cursor.execute("insert into TestRowids (IntCol) values (1)")
        self.connection.commit()
        self.cursor.execute("select * from TestRowids")
        self.assertEqual(self.cursor.fetchone(), (1, None, None))

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# conversions.pyx
#
# Cython file defining the conversions between data types that are supported by
# the thin client (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef converter_dict = {
    (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_NUMBER): float,
    (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_NUMBER): float,
    (TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_NUMBER): float,
    (TNS_DATA_TYPE_BINARY_INTEGER, TNS_DATA_TYPE_NUMBER):float,
    (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_BINARY_INTEGER):_to_binary_int,
    (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_BINARY_INTEGER): _to_binary_int,
    (TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_BINARY_INTEGER): _to_binary_int,
    (TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_BINARY_INTEGER): _to_binary_int,
    (TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_BINARY_INTEGER):
            _to_binary_int,
    (TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_BINARY_INTEGER):
            _to_binary_int,
    (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_VARCHAR): NUM_TYPE_STR,
    (TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_CHAR): NUM_TYPE_STR,
    (TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_LONG): NUM_TYPE_STR,
    (TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_ROWID, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_ROWID, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_ROWID, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_INTERVAL_DS, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_INTERVAL_DS, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_INTERVAL_DS, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_BINARY_INTEGER, TNS_DATA_TYPE_VARCHAR): str,
    (TNS_DATA_TYPE_BINARY_INTEGER, TNS_DATA_TYPE_CHAR): str,
    (TNS_DATA_TYPE_BINARY_INTEGER, TNS_DATA_TYPE_LONG): str,
    (TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_DATE): _tstamp_to_date,
    (TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_DATE): _tstamp_to_date,
    (TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_DATE): _tstamp_to_date,
    (TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_BINARY_DOUBLE): NUM_TYPE_FLOAT,
    (TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_BINARY_DOUBLE): float,
    (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_BINARY_DOUBLE): float,
    (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_BINARY_DOUBLE): float,
    (TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_BINARY_DOUBLE): float,
    (TNS_DATA_TYPE_NUMBER, TNS_DATA_TYPE_BINARY_FLOAT): NUM_TYPE_FLOAT,
    (TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_BINARY_FLOAT): float,
    (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_BINARY_FLOAT): float,
    (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_BINARY_FLOAT): float,
    (TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_BINARY_FLOAT): float,
    (TNS_DATA_TYPE_BINARY_FLOAT, TNS_DATA_TYPE_NUMBER): float,
    (TNS_DATA_TYPE_BINARY_DOUBLE, TNS_DATA_TYPE_NUMBER): float,
    (TNS_DATA_TYPE_BLOB, TNS_DATA_TYPE_LONG_RAW): TNS_DATA_TYPE_LONG_RAW,
    (TNS_DATA_TYPE_BLOB, TNS_DATA_TYPE_RAW): TNS_DATA_TYPE_LONG_RAW,
    (TNS_DATA_TYPE_CLOB, TNS_DATA_TYPE_CHAR): TNS_DATA_TYPE_LONG,
    (TNS_DATA_TYPE_CLOB, TNS_DATA_TYPE_VARCHAR): TNS_DATA_TYPE_LONG,
    (TNS_DATA_TYPE_CLOB, TNS_DATA_TYPE_LONG): TNS_DATA_TYPE_LONG,
    (TNS_DATA_TYPE_JSON, TNS_DATA_TYPE_VARCHAR): TNS_DATA_TYPE_LONG,
    (TNS_DATA_TYPE_JSON, TNS_DATA_TYPE_CHAR): TNS_DATA_TYPE_LONG,
    (TNS_DATA_TYPE_JSON, TNS_DATA_TYPE_RAW): TNS_DATA_TYPE_LONG_RAW,
    (TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_TIMESTAMP_LTZ): None,
    (TNS_DATA_TYPE_TIMESTAMP_TZ, TNS_DATA_TYPE_TIMESTAMP): None,
    (TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_TIMESTAMP_TZ): None,
    (TNS_DATA_TYPE_TIMESTAMP_LTZ, TNS_DATA_TYPE_TIMESTAMP): None,
    (TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_TIMESTAMP_LTZ): None,
    (TNS_DATA_TYPE_TIMESTAMP, TNS_DATA_TYPE_TIMESTAMP_TZ): None,
    (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_TIMESTAMP_LTZ): None,
    (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_TIMESTAMP): None,
    (TNS_DATA_TYPE_DATE, TNS_DATA_TYPE_TIMESTAMP_TZ): None,
    (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_VARCHAR): None,
    (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_CHAR): None,
    (TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_VARCHAR): None,
    (TNS_DATA_TYPE_LONG, TNS_DATA_TYPE_CHAR): None,
    (TNS_DATA_TYPE_VARCHAR, TNS_DATA_TYPE_LONG): None,
    (TNS_DATA_TYPE_CHAR, TNS_DATA_TYPE_LONG): None,
}

cdef object _to_binary_int(object fetch_value):
    return int(PY_TYPE_DECIMAL(fetch_value))

cdef object _tstamp_to_date(object fetch_value):
    return fetch_value.replace(microsecond=0)

cdef int conversion_helper(ThinVarImpl output_var, FetchInfo fetch_info,
                           bint *requires_define) except -1:
    cdef:
        uint8_t fetch_ora_type_num, output_ora_type_num, csfrm
        object key, value

    fetch_ora_type_num = fetch_info._dbtype._ora_type_num
    output_ora_type_num = output_var.dbtype._ora_type_num

    key = (fetch_ora_type_num, output_ora_type_num)
    try:
        value = converter_dict[key]
        if isinstance(value, int):
            if fetch_ora_type_num == TNS_DATA_TYPE_NUMBER:
                output_var._preferred_num_type = value
            else:
                requires_define[0] = True
                csfrm = output_var.dbtype._csfrm
                fetch_info._dbtype = DbType._from_ora_type_and_csfrm(value,
                                                                     csfrm)
        else:
            output_var._conv_func = value
    except:
        errors._raise_err(errors.ERR_INCONSISTENT_DATATYPES,
                          input_type=fetch_info._dbtype.name,
                          output_type=output_var.dbtype.name)
